# All Rights Reserved 2018
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import operator
from unittest import mock

from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from neutronclient.tests.unit.osc.v2 import fakes as test_fakes

from networking_l2gw.l2gatewayclient.osc import l2gw as osc_l2gw
from networking_l2gw.tests.unit.l2gatewayclient.osc import fakes


columns_long = tuple(col for col, _, listing_mode in osc_l2gw._attr_map
                     if listing_mode in (column_util.LIST_BOTH,
                                         column_util.LIST_LONG_ONLY))
headers_long = tuple(head for _, head, listing_mode in osc_l2gw._attr_map
                     if listing_mode in (column_util.LIST_BOTH,
                                         column_util.LIST_LONG_ONLY))
sorted_attr_map = sorted(osc_l2gw._attr_map, key=operator.itemgetter(1))
sorted_columns = tuple(col for col, _, _ in sorted_attr_map)
sorted_headers = tuple(head for _, head, _ in sorted_attr_map)


def _get_data(attrs, columns=sorted_columns):
    return osc_utils.get_dict_properties(attrs, columns,
                                         formatters=osc_l2gw._formatters)


class TestCreateL2gw(test_fakes.TestNeutronClientOSCV2):

    columns = (
        'Devices',
        'ID',
        'Name',
        'Tenant'
    )

    def setUp(self):
        super(TestCreateL2gw, self).setUp()
        self.cmd = osc_l2gw.CreateL2gw(self.app, self.namespace)

    def _assert_create_succeeded(self, fake_l2gw, arg_list, verify_list):
        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.neutronclient.post.assert_called_once_with(
            osc_l2gw.object_path,
            {osc_l2gw.L2_GATEWAY:
                {'name': fake_l2gw['name'], 'devices': fake_l2gw['devices']}}
        )
        self.assertEqual(self.columns, columns)
        self.assertItemEqual(_get_data(fake_l2gw), data)

    def test_create_l2gw(self):
        """Test Create l2gateway."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()

        self.neutronclient.post = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: fake_l2gw})
        l2gw_device = fake_l2gw['devices'][0]
        arg_list = [
            '--device', 'name=' + l2gw_device['device_name'] +
            ',interface_names=' + l2gw_device['interfaces'][0]['name'],
            fake_l2gw['name']
        ]

        verify_list = [
            ('devices', [
                {'interface_names': l2gw_device['interfaces'][0]['name'],
                 'name': l2gw_device['device_name']}]),
            ('name', fake_l2gw['name']),
        ]

        self._assert_create_succeeded(fake_l2gw, arg_list, verify_list)

    def test_create_l2gateway_with_multiple_devices(self):
        """Test Create l2gateway for multiple devices."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw(num_dev=2)

        self.neutronclient.post = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: fake_l2gw})
        l2gw_device_1 = fake_l2gw['devices'][0]
        l2gw_device_2 = fake_l2gw['devices'][1]
        arg_list = [
            '--device', 'name=' + l2gw_device_1['device_name'] +
            ',interface_names=' +
            l2gw_device_1['interfaces'][0]['name'],
            '--device', 'name=' + l2gw_device_2['device_name'] +
            ',interface_names=' +
            l2gw_device_2['interfaces'][0]['name'],
            fake_l2gw['name']
        ]
        verify_list = [
            ('devices', [
                {'interface_names': l2gw_device_1['interfaces'][0]['name'],
                 'name': l2gw_device_1['device_name']},
                {'interface_names': l2gw_device_2['interfaces'][0]['name'],
                 'name': l2gw_device_2['device_name']},
            ]),
            ('name', fake_l2gw['name']),
        ]

        self._assert_create_succeeded(fake_l2gw, arg_list, verify_list)

    def test_create_l2gateway_with_multiple_interfaces(self):
        """Test Create l2gateway with multiple interfaces."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw(num_if=2)

        self.neutronclient.post = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: fake_l2gw})
        l2gw_device = fake_l2gw['devices'][0]
        l2gw_interface_1 = l2gw_device['interfaces'][0]
        l2gw_interface_2 = l2gw_device['interfaces'][1]
        arg_list = [
            '--device', 'name=' + l2gw_device['device_name'] +
            ',interface_names=' + l2gw_interface_1['name'] + ';' +
            l2gw_interface_2['name'],
            fake_l2gw['name']
        ]
        verify_list = [
            ('devices', [
                {
                    'interface_names':
                        l2gw_interface_1['name'] + ';' +
                        l2gw_interface_2['name'],
                    'name': l2gw_device['device_name']
                }
            ]),
            ('name', fake_l2gw['name']),
        ]

        self._assert_create_succeeded(fake_l2gw, arg_list, verify_list)

    def test_create_l2gateway_with_segmentation_id(self):
        """Test Create l2gateway with segmentation-id."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        fake_l2gw['devices'][0]['interfaces'][0]['segmentation_id'] = ['42']

        self.neutronclient.post = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: fake_l2gw})
        l2gw_device = fake_l2gw['devices'][0]
        l2gw_interface = l2gw_device['interfaces'][0]
        arg_list = [
            '--device', 'name=' + l2gw_device['device_name'] +
            ',interface_names=' + l2gw_interface['name'] + '|' +
            '#'.join(l2gw_interface['segmentation_id']),
            fake_l2gw['name']
        ]

        verify_list = [
            ('devices', [
                {'interface_names': l2gw_device['interfaces'][0]['name'] +
                 '|' + '#'.join(l2gw_interface['segmentation_id']),
                 'name': l2gw_device['device_name']}]),
            ('name', fake_l2gw['name']),
        ]

        self._assert_create_succeeded(fake_l2gw, arg_list, verify_list)

    def test_create_l2gateway_with_mul_segmentation_id(self):
        """Test Create l2gateway with multiple segmentation-ids."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        fake_l2gw['devices'][0]['interfaces'][0]['segmentation_id'] = ['42',
                                                                       '43']

        self.neutronclient.post = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: fake_l2gw})
        l2gw_device = fake_l2gw['devices'][0]
        l2gw_interface = l2gw_device['interfaces'][0]
        arg_list = [
            '--device', 'name=' + l2gw_device['device_name'] +
            ',interface_names=' + l2gw_interface['name'] + '|' +
            '#'.join(l2gw_interface['segmentation_id']),
            fake_l2gw['name'],
        ]

        verify_list = [
            ('devices', [
                {'interface_names': l2gw_device['interfaces'][0]['name'] +
                 '|' + '#'.join(l2gw_interface['segmentation_id']),
                 'name': l2gw_device['device_name']}]),
            ('name', fake_l2gw['name']),
        ]

        self._assert_create_succeeded(fake_l2gw, arg_list, verify_list)


class TestListL2gw(test_fakes.TestNeutronClientOSCV2):
    def setUp(self):
        super(TestListL2gw, self).setUp()
        self.cmd = osc_l2gw.ListL2gw(self.app, self.namespace)

    def test_list_l2gateway(self):
        """Test List l2gateways."""

        fake_l2gws = fakes.FakeL2GW.create_l2gws(count=4)
        self.neutronclient.list = mock.Mock(return_value=fake_l2gws)
        arg_list = []
        verify_list = []

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        headers, data = self.cmd.take_action(parsed_args)

        self.neutronclient.list.assert_called_once()
        self.assertEqual(headers, list(headers_long))
        self.assertListItemEqual(
            list(data),
            [_get_data(fake_l2gw, columns_long) for fake_l2gw
             in fake_l2gws[osc_l2gw.L2_GATEWAYS]]
        )


class TestDeleteL2gw(test_fakes.TestNeutronClientOSCV2):
    def setUp(self):
        super(TestDeleteL2gw, self).setUp()
        self.neutronclient.find_resource = mock.Mock(
            side_effect=lambda _, name_or_id: {'id': name_or_id})
        self.cmd = osc_l2gw.DeleteL2gw(self.app, self.namespace)

    def test_delete_l2gateway(self):
        """Test Delete l2gateway."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        self.neutronclient.delete = mock.Mock()

        arg_list = [
            fake_l2gw['id'],
        ]
        verify_list = [
            (osc_l2gw.L2_GATEWAYS, [fake_l2gw['id']]),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        result = self.cmd.take_action(parsed_args)

        self.neutronclient.delete.assert_called_once_with(
            osc_l2gw.resource_path % fake_l2gw['id'])
        self.assertIsNone(result)


class TestShowL2gw(test_fakes.TestNeutronClientOSCV2):
    def setUp(self):
        super(TestShowL2gw, self).setUp()
        self.neutronclient.find_resource = mock.Mock(
            side_effect=lambda _, name_or_id: {'id': name_or_id})
        self.cmd = osc_l2gw.ShowL2gw(self.app, self.namespace)

    def test_show_l2gateway(self):
        """Test Show l2gateway: --fields id --fields name myid."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        self.neutronclient.get = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: fake_l2gw})
        arg_list = [
            fake_l2gw['id'],
        ]
        verify_list = [
            (osc_l2gw.L2_GATEWAY, fake_l2gw['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        headers, data = self.cmd.take_action(parsed_args)

        self.neutronclient.get.assert_called_once_with(
            osc_l2gw.resource_path % fake_l2gw['id'])
        self.assertEqual(sorted_headers, headers)
        self.assertItemEqual(_get_data(fake_l2gw), data)


class TestUpdateL2gw(test_fakes.TestNeutronClientOSCV2):
    _new_device_name = 'new_device'
    _new_interface = 'new_interface'
    _new_name = 'new_name'

    columns = (
        'Devices',
        'ID',
        'Name',
        'Tenant'
    )

    def setUp(self):
        super(TestUpdateL2gw, self).setUp()
        self.cmd = osc_l2gw.UpdateL2gw(self.app, self.namespace)
        self.neutronclient.find_resource = mock.Mock(
            side_effect=lambda _, name_or_id: {'id': name_or_id})

    def _assert_update_succeeded(self, new_l2gw, attrs, columns, data):
        self.neutronclient.put.assert_called_once_with(
            osc_l2gw.resource_path % new_l2gw['id'],
            {osc_l2gw.L2_GATEWAY: attrs})
        self.assertEqual(self.columns, columns)
        self.assertItemEqual(_get_data(new_l2gw), data)

    def test_update_l2gateway(self):
        """Test Update l2gateway."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        new_l2gw = copy.deepcopy(fake_l2gw)
        new_l2gw['name'] = self._new_name
        new_l2gw['devices'][0]['device_name'] = self._new_device_name
        new_l2gw['devices'][0]['interfaces'][0]['name'] = self._new_interface

        self.neutronclient.put = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: new_l2gw})

        arg_list = [
            fake_l2gw['id'],
            '--name', self._new_name,
            '--device', 'name=' + self._new_device_name +
                        ',interface_names=' + self._new_interface,
        ]
        verify_list = [
            ('name', self._new_name),
            ('devices', [
                {'interface_names': self._new_interface,
                 'name': self._new_device_name}]),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            'name': self._new_name,
            'devices': [
                {'interfaces': [
                    {'name': self._new_interface}],
                 'device_name': self._new_device_name}
            ]
        }

        self._assert_update_succeeded(new_l2gw, attrs, columns, data)

    def test_update_l2gateway_name(self):
        """Test Update l2gateway name."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        new_l2gw = copy.deepcopy(fake_l2gw)
        new_l2gw['name'] = self._new_name

        self.neutronclient.put = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: new_l2gw})

        arg_list = [
            fake_l2gw['id'],
            '--name', self._new_name,
        ]
        verify_list = [('name', self._new_name)]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {'name': self._new_name}

        self._assert_update_succeeded(new_l2gw, attrs, columns, data)

    def test_update_l2gateway_with_multiple_interfaces(self):
        """Test Update l2gateway with multiple interfaces."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()
        new_l2gw = copy.deepcopy(fake_l2gw)
        new_l2gw['devices'][0]['interfaces'].append(
            {'name': self._new_interface})

        self.neutronclient.put = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: new_l2gw})

        l2gw_device = new_l2gw['devices'][0]
        l2gw_interface_1 = l2gw_device['interfaces'][0]
        l2gw_interface_2 = l2gw_device['interfaces'][1]
        arg_list = [
            fake_l2gw['id'],
            '--device', 'name=' + l2gw_device['device_name'] +
                        ',interface_names=' + l2gw_interface_1['name'] + ';' +
                        l2gw_interface_2['name']
        ]

        verify_list = [
            ('devices', [
                {
                    'interface_names':
                        l2gw_interface_1['name'] + ';' +
                        l2gw_interface_2['name'],
                    'name': l2gw_device['device_name']
                }
            ]),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            'devices': [
                {'device_name': l2gw_device['device_name'],
                 'interfaces': [
                     {'name': l2gw_interface_1['name']},
                     {'name': self._new_interface}]}
            ]
        }

        self._assert_update_succeeded(new_l2gw, attrs, columns, data)

    def test_update_l2gateway_with_segmentation_id(self):
        """Test Update l2gateway with segmentation-id."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()

        new_l2gw = copy.deepcopy(fake_l2gw)
        new_l2gw['devices'][0]['interfaces'][0]['segmentation_id'] = ['42']

        self.neutronclient.put = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: new_l2gw})

        l2gw_device = new_l2gw['devices'][0]
        l2gw_interface = l2gw_device['interfaces'][0]
        arg_list = [
            fake_l2gw['id'],
            '--device', 'name=' + l2gw_device['device_name'] +
            ',interface_names=' + l2gw_interface['name'] + '|' +
            '#'.join(l2gw_interface['segmentation_id']),
        ]

        verify_list = [
            ('devices', [
                {'interface_names': l2gw_device['interfaces'][0]['name'] +
                 '|' + '#'.join(l2gw_interface['segmentation_id']),
                 'name': l2gw_device['device_name']}]),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            'devices': [
                {'device_name': l2gw_device['device_name'],
                 'interfaces': [
                     {'name': l2gw_interface['name'],
                      'segmentation_id': l2gw_interface['segmentation_id']}]
                 }
            ]
        }

        self._assert_update_succeeded(new_l2gw, attrs, columns, data)

    def test_update_l2gateway_with_mul_segmentation_ids(self):
        """Test Update l2gateway with multiple segmentation-ids."""

        fake_l2gw = fakes.FakeL2GW.create_l2gw()

        new_l2gw = copy.deepcopy(fake_l2gw)
        new_l2gw['devices'][0]['interfaces'][0]['segmentation_id'] = ['42',
                                                                      '43']
        self.neutronclient.put = mock.Mock(
            return_value={osc_l2gw.L2_GATEWAY: new_l2gw})

        l2gw_device = new_l2gw['devices'][0]
        l2gw_interface = l2gw_device['interfaces'][0]
        arg_list = [
            fake_l2gw['id'],
            '--device', 'name=' + l2gw_device['device_name'] +
            ',interface_names=' + l2gw_interface['name'] + '|' +
            '#'.join(l2gw_interface['segmentation_id']),
        ]

        verify_list = [
            ('devices', [
                {'interface_names': l2gw_device['interfaces'][0]['name'] +
                 '|' + '#'.join(l2gw_interface['segmentation_id']),
                 'name': l2gw_device['device_name']}]),
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            'devices': [
                {'device_name': l2gw_device['device_name'],
                 'interfaces': [
                     {'name': l2gw_interface['name'],
                      'segmentation_id': l2gw_interface['segmentation_id']}]
                 }
            ]
        }

        self._assert_update_succeeded(new_l2gw, attrs, columns, data)



# Copyright 2015 OpenStack Foundation
# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import abc

from neutron_lib.api import extensions

from neutron.api.v2 import resource_helper

from networking_l2gw.services.l2gateway.common import constants

RESOURCE_ATTRIBUTE_MAP = {
    constants.L2_GATEWAYS_CONNECTION: {
        'id': {'allow_post': False, 'allow_put': False,
               'is_visible': True},
        'l2_gateway_id': {'allow_post': True, 'allow_put': False,
                          'validate': {'type:string': None},
                          'is_visible': True, 'default': ''},
        'network_id': {'allow_post': True, 'allow_put': False,
                       'validate': {'type:string': None},
                       'is_visible': True},
        'segmentation_id': {'allow_post': True, 'allow_put': False,
                            'validate': {'type:string': None},
                            'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True,
                      'is_visible': True}
    },
}


class L2gatewayconnection(extensions.ExtensionDescriptor):

    """API extension for Layer-2 Gateway  connection support."""

    @classmethod
    def get_name(cls):
        return "L2 Gateway connection"

    @classmethod
    def get_alias(cls):
        return "l2-gateway-connection"

    @classmethod
    def get_description(cls):
        return "Connects Neutron networks with external networks at layer 2."

    @classmethod
    def get_updated(cls):
        return "2014-01-01T00:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""

        mem_actions = {}
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        resources = resource_helper.build_resource_info(plural_mappings,
                                                        RESOURCE_ATTRIBUTE_MAP,
                                                        constants.L2GW,
                                                        action_map=mem_actions,
                                                        register_quota=True,
                                                        translate_name=True)

        return resources

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


class L2GatewayConnectionPluginBase(object):

    @abc.abstractmethod
    def delete_l2_gateway_connection(self, context, l2_gateway_id,
                                     network_mapping_list):
        pass

    @abc.abstractmethod
    def create_l2_gateway_connection(self, context, l2_gateway_id,
                                     network_mapping_list):
        pass

    @abc.abstractmethod
    def get_l2_gateway_connections(self, context, filters=None,
                                   fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        pass

    @abc.abstractmethod
    def get_l2_gateway_connection(self, context, id, fields=None):
        pass



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""DB_Models_for_OVSDB_Hardware_VTEP_Schema

Revision ID: 54c9c8fe22bf
Revises: 42438454c556
Create Date: 2015-01-27 02:05:21.599215

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '54c9c8fe22bf'
down_revision = 'start_networking_l2gw'


def upgrade():
    op.create_table('physical_locators',
                    sa.Column('dst_ip', sa.String(length=64), nullable=True),
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'ovsdb_identifier'))

    op.create_table('physical_switches',
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('tunnel_ip', sa.String(length=64),
                              nullable=True),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.Column('switch_fault_status', sa.String(length=32),
                              nullable=True),
                    sa.PrimaryKeyConstraint('uuid', 'ovsdb_identifier'))

    op.create_table('physical_ports',
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('physical_switch_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.Column('port_fault_status', sa.String(length=32),
                              nullable=True),
                    sa.PrimaryKeyConstraint('uuid', 'ovsdb_identifier'))

    op.create_table('logical_switches',
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('key', sa.Integer(), nullable=True),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'ovsdb_identifier'))

    op.create_table('ucast_macs_locals',
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('mac', sa.String(length=32), nullable=True),
                    sa.Column('logical_switch_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('physical_locator_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('ip_address', sa.String(length=64),
                              nullable=True),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'ovsdb_identifier'))

    op.create_table('ucast_macs_remotes',
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('mac', sa.String(length=32), nullable=True),
                    sa.Column('logical_switch_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('physical_locator_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('ip_address', sa.String(length=64),
                              nullable=True),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'ovsdb_identifier'))

    op.create_table('vlan_bindings',
                    sa.Column('port_uuid', sa.String(length=36),
                              nullable=False),
                    sa.Column('vlan', sa.Integer(), nullable=False),
                    sa.Column('logical_switch_uuid', sa.String(length=36),
                              nullable=False),
                    sa.Column('ovsdb_identifier', sa.String(length=64),
                              nullable=False),
                    sa.PrimaryKeyConstraint('port_uuid', 'ovsdb_identifier',
                                            'vlan', 'logical_switch_uuid'))



# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import exceptions

from neutron.tests import base

from networking_l2gw.services.l2gateway.common import l2gw_validators


class TestL2gwValidators(base.BaseTestCase):

    def test_validate_gwdevice_list(self):
        ret_msg = l2gw_validators.validate_gwdevice_list(None)
        msg = "Cannot create a gateway with an empty device list"
        self.assertEqual(msg, ret_msg)

    def test_for_empty_device_name(self):
        test_device = [{"interfaces": "port"}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "Cannot create a gateway with an empty device_name"
        self.assertEqual(msg, ret_msg)

    def test_for_empty_interface(self):
        test_device = [{"device_name": "switch1"}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "Cannot create a gateway with an empty interfaces"
        self.assertEqual(msg, ret_msg)

    def test_for_interfaces_type_list_of_dicts(self):
        test_device = [{"device_name": "switch1", "interfaces": "port"}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "interfaces format is not a type list of dicts"
        self.assertEqual(msg, ret_msg)

    def test_for_interface_type_dict(self):
        test_device = [{"device_name": "switch1", "interfaces": ["port"]}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "interfaces format is not a type dict"
        self.assertEqual(msg, ret_msg)

    def test_for_empty_interface_name(self):
        test_device = [{"device_name": "switch1",
                        "interfaces": [{'name': ""}]}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "Cannot create a gateway with an empty interface name"
        self.assertEqual(msg, ret_msg)

    def test_segmentation_id_in_interface(self):
        test_device = [{"device_name": "switch1",
                        "interfaces": [{'name': "i1", 'segmentation_id': ''}]}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "segmentation_id_list should not be empty"
        self.assertEqual(msg, ret_msg)

    def test_segmentation_id_type(self):
        test_device = [{"device_name": "switch1",
                        "interfaces": [{'name': "i1",
                                        'segmentation_id': '67'}]}]
        ret_msg = l2gw_validators.validate_gwdevice_list(test_device)
        msg = "segmentation_id type should be of list type "
        self.assertEqual(msg, ret_msg)

    def test_validate_network_mapping_list_with_seg_id(self):
        network_mapping = {'segmentation_id': 67}
        check_vlan = True
        self.assertRaises(exceptions.InvalidInput,
                          l2gw_validators.validate_network_mapping_list,
                          network_mapping, check_vlan)

    def test_validate_network_mapping_list_without_seg_id(self):
        network_mapping = {'segmentation_id': ''}
        check_vlan = False
        self.assertRaises(exceptions.InvalidInput,
                          l2gw_validators.validate_network_mapping_list,
                          network_mapping, check_vlan)

    def test_validate_network_mapping_list_for_empty_network_id(self):
        network_mapping = {'segmentation_id': '', 'network_id': ''}
        check_vlan = True
        self.assertRaises(exceptions.InvalidInput,
                          l2gw_validators.validate_network_mapping_list,
                          network_mapping, check_vlan)

    def test_is_valid_vlan_id_for_non_integer_value(self):
        seg_id = 'a'
        self.assertRaises(exceptions.InvalidInput,
                          l2gw_validators.is_valid_vlan_id,
                          seg_id)

    def test_is_valid_vlan_id_for_value_less_than_0(self):
        seg_id = -1
        self.assertRaises(exceptions.InvalidInput,
                          l2gw_validators.is_valid_vlan_id,
                          seg_id)

    def test_is_valid_vlan_id_for_value_greater_than_4095(self):
        seg_id = 4096
        self.assertRaises(exceptions.InvalidInput,
                          l2gw_validators.is_valid_vlan_id,
                          seg_id)



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests import base

from networking_l2gw.services.l2gateway.agent import agent_api


class L2GatewayAgentApiTestCase(base.BaseTestCase):

    def setUp(self):
        self.client_mock_p = mock.patch.object(agent_api.n_rpc, 'get_client')
        self.client_mock = self.client_mock_p.start()
        self.topic = 'foo_topic'
        self.host = 'foo_host'

        self.agent_rpc = agent_api.L2GatewayAgentApi(
            self.topic, self.host)
        super(L2GatewayAgentApiTestCase, self).setUp()

    def test_update_ovsdb_changes(self):
        cctxt = mock.Mock()
        context = mock.Mock()
        fake_activity = 1
        self.agent_rpc.client.prepare.return_value = cctxt
        self.agent_rpc.update_ovsdb_changes(context, fake_activity, mock.ANY)
        cctxt.cast.assert_called_with(
            context, 'update_ovsdb_changes',
            activity=fake_activity, ovsdb_data=mock.ANY)

    def test_notify_ovsdb_states(self):
        cctxt = mock.Mock()
        context = mock.Mock()
        self.agent_rpc.client.prepare.return_value = cctxt
        self.agent_rpc.notify_ovsdb_states(context, mock.ANY)
        cctxt.cast.assert_called_with(
            context, 'notify_ovsdb_states', ovsdb_states=mock.ANY)



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
import socket
from unittest import mock

import eventlet

from neutron.agent import rpc as agent_rpc
from neutron.conf.agent import common as agent_config
from neutron.tests import base
from neutron_lib import context
from neutron_lib import rpc

from oslo_config import cfg
from oslo_service import loopingcall

from networking_l2gw.services.l2gateway.agent import agent_api
from networking_l2gw.services.l2gateway.agent import base_agent_manager
from networking_l2gw.services.l2gateway.agent import l2gateway_config
from networking_l2gw.services.l2gateway.agent.ovsdb import manager
from networking_l2gw.services.l2gateway.agent.ovsdb import ovsdb_common_class
from networking_l2gw.services.l2gateway.agent.ovsdb import ovsdb_monitor
from networking_l2gw.services.l2gateway.agent.ovsdb import ovsdb_writer
from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway.common import constants as n_const


class TestManager(base.BaseTestCase):

    def setUp(self):
        super(TestManager, self).setUp()
        self.conf = cfg.CONF
        config.register_ovsdb_opts_helper(self.conf)
        agent_config.register_agent_state_opts_helper(self.conf)
        self.driver_mock = mock.Mock()
        self.fake_record_dict = {n_const.OVSDB_IDENTIFIER: 'fake_ovsdb_id'}
        cfg.CONF.set_override('report_interval', 1, 'AGENT')
        self.plugin_rpc = mock.patch.object(agent_api,
                                            'L2GatewayAgentApi').start()
        self.context = mock.Mock
        self.cntxt = mock.patch.object(context,
                                       'get_admin_context_without_session'
                                       ).start()
        self.test_rpc = mock.patch.object(rpc, 'get_client').start()
        self.mock_looping_call = mock.patch.object(loopingcall,
                                                   'FixedIntervalLoopingCall'
                                                   ).start()
        self.l2gw_agent_manager = manager.OVSDBManager(
            self.conf)
        self.l2gw_agent_manager.plugin_rpc = self.plugin_rpc
        self.fake_config_json = {n_const.OVSDB_IDENTIFIER:
                                 'fake_ovsdb_identifier',
                                 'ovsdb_ip': '5.5.5.5',
                                 'ovsdb_port': '6672',
                                 'private_key': 'dummy_key',
                                 'enable_ssl': False,
                                 'certificate': 'dummy_cert',
                                 'ca_cert': 'dummy_ca'}

    def test_process_ovsdb_host(self):
        fake_host = "ovsdb1:10.10.10.10:6632"
        cfg.CONF.set_override('l2_gw_agent_priv_key_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_cert_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_ca_cert_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        with mock.patch.object(os.path, 'isfile',
                               return_value=True) as mock_isfile, \
                mock.patch.object(manager.LOG, 'exception') as mock_log_exc:
            self.l2gw_agent_manager._process_ovsdb_host(fake_host)
            self.assertTrue(mock_isfile.called)
            self.assertFalse(mock_log_exc.called)

    def test_process_ovsdb_host_for_certs_not_found(self):
        fake_host = "ovsdb1:10.10.10.10:6632"
        cfg.CONF.set_override('l2_gw_agent_priv_key_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_cert_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_ca_cert_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        with mock.patch.object(os.path, 'isfile',
                               return_value=False) as mock_isfile, \
                mock.patch.object(manager.LOG, 'exception') as mock_log_exc:
            self.l2gw_agent_manager._process_ovsdb_host(fake_host)
            self.assertTrue(mock_isfile.called)
            self.assertTrue(mock_log_exc.called)

    def test_extract_ovsdb_config(self):
        fake_ovsdb_config = {n_const.OVSDB_IDENTIFIER: 'host2',
                             'ovsdb_ip': '10.10.10.10',
                             'ovsdb_port': '6632',
                             'private_key': '/home/someuser/fakedir/host2.key',
                             'use_ssl': True,
                             'certificate':
                             '/home/someuser/fakedir/host2.cert',
                             'ca_cert': '/home/someuser/fakedir/host2.ca_cert'}
        cfg.CONF.set_override('ovsdb_hosts',
                              'host2:10.10.10.10:6632',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_priv_key_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_cert_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        cfg.CONF.set_override('l2_gw_agent_ca_cert_base_path',
                              '/home/someuser/fakedir',
                              'ovsdb')
        self.l2gw_agent_manager._extract_ovsdb_config(cfg.CONF)
        l2gwconfig = l2gateway_config.L2GatewayConfig(fake_ovsdb_config)
        gw = self.l2gw_agent_manager.gateways.get(
            fake_ovsdb_config[n_const.OVSDB_IDENTIFIER])
        self.assertEqual(l2gwconfig.ovsdb_identifier, gw.ovsdb_identifier)
        self.assertEqual(l2gwconfig.use_ssl, gw.use_ssl)
        self.assertEqual(l2gwconfig.ovsdb_ip, gw.ovsdb_ip)
        self.assertEqual(l2gwconfig.ovsdb_port, gw.ovsdb_port)
        self.assertEqual(l2gwconfig.private_key, gw.private_key)
        self.assertEqual(l2gwconfig.certificate, gw.certificate)
        self.assertEqual(l2gwconfig.ca_cert, gw.ca_cert)
        cfg.CONF.set_override('max_connection_retries',
                              25,
                              'ovsdb')
        self.assertRaises(SystemExit,
                          self.l2gw_agent_manager.
                          _extract_ovsdb_config,
                          cfg.CONF)

    def test_connect_to_ovsdb_server(self):
        self.l2gw_agent_manager.gateways = {}
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        gateway = l2gateway_config.L2GatewayConfig(self.fake_config_json)
        ovsdb_ident = self.fake_config_json.get(n_const.OVSDB_IDENTIFIER)
        self.l2gw_agent_manager.gateways[ovsdb_ident] = gateway
        with mock.patch.object(ovsdb_monitor,
                               'OVSDBMonitor') as ovsdb_connection, \
                mock.patch.object(eventlet.greenthread,
                                  'spawn_n') as event_spawn, \
                mock.patch.object(manager.OVSDBManager,
                                  'agent_to_plugin_rpc') as call_back, \
                mock.patch.object(self.plugin_rpc,
                                  'notify_ovsdb_states') as notify:
            self.l2gw_agent_manager._connect_to_ovsdb_server()
            self.assertTrue(event_spawn.called)
            self.assertTrue(ovsdb_connection.called)
            ovsdb_connection.assert_called_with(
                self.conf.ovsdb, gateway, call_back)
            notify.assert_called_once_with(mock.ANY, mock.ANY)

    def test_connect_to_ovsdb_server_with_exc(self):
        self.l2gw_agent_manager.gateways = {}
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        gateway = l2gateway_config.L2GatewayConfig(self.fake_config_json)
        ovsdb_ident = self.fake_config_json.get(n_const.OVSDB_IDENTIFIER)
        self.l2gw_agent_manager.gateways[ovsdb_ident] = gateway
        with mock.patch.object(ovsdb_monitor.OVSDBMonitor,
                               '__init__',
                               side_effect=socket.error
                               ), \
                mock.patch.object(eventlet.greenthread,
                                  'spawn_n') as event_spawn, \
                mock.patch.object(manager.LOG, 'error'):
            self.l2gw_agent_manager._connect_to_ovsdb_server()
            event_spawn.assert_not_called()

    def test_handle_report_state_failure(self):
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        with mock.patch.object(self.l2gw_agent_manager,
                               '_disconnect_all_ovsdb_servers'
                               ) as mock_disconnect_ovsdb_servers, \
                mock.patch.object(agent_rpc.PluginReportStateAPI,
                                  'report_state',
                                  side_effect=Exception), \
                mock.patch.object(base_agent_manager.LOG,
                                  'exception'), \
                mock.patch.object(self.l2gw_agent_manager,
                                  '_stop_looping_task') as mock_stop_looping:
            self.l2gw_agent_manager._report_state()
            self.assertEqual(self.l2gw_agent_manager.l2gw_agent_type,
                             '')
            self.assertEqual(self.l2gw_agent_manager.
                             agent_state.get('configurations'
                                             )[n_const.L2GW_AGENT_TYPE
                                               ],
                             '')
            self.assertTrue(mock_disconnect_ovsdb_servers.called)
            self.assertTrue(mock_stop_looping.called)

    def test_is_valid_request_fails(self):
        self.l2gw_agent_manager.gateways = {}
        fake_ovsdb_identifier = 'fake_ovsdb_identifier_2'
        gateway = l2gateway_config.L2GatewayConfig(self.fake_config_json)
        self.l2gw_agent_manager.gateways['fake_ovsdb_identifier'] = gateway
        with mock.patch.object(manager.LOG,
                               'warning') as logger_call:
            self.l2gw_agent_manager._is_valid_request(
                fake_ovsdb_identifier)
            self.assertEqual(1, logger_call.call_count)

    def test_open_connection(self):
        self.l2gw_agent_manager.gateways = {}
        fake_ovsdb_identifier = 'fake_ovsdb_identifier'
        gateway = l2gateway_config.L2GatewayConfig(self.fake_config_json)
        self.l2gw_agent_manager.gateways['fake_ovsdb_identifier'] = gateway
        with mock.patch.object(manager.LOG,
                               'warning') as logger_call:
            with mock.patch.object(ovsdb_writer,
                                   'OVSDBWriter') as ovsdb_connection:
                is_valid_request = self.l2gw_agent_manager._is_valid_request(
                    fake_ovsdb_identifier)
                with self.l2gw_agent_manager._open_connection(
                        fake_ovsdb_identifier):
                    self.assertTrue(is_valid_request)
                    self.assertEqual(0, logger_call.call_count)
                    self.assertTrue(ovsdb_connection.called)

    def test_open_connection_with_socket_error(self):
        self.l2gw_agent_manager.gateways = {}
        gateway = l2gateway_config.L2GatewayConfig(self.fake_config_json)
        self.l2gw_agent_manager.gateways['fake_ovsdb_identifier'] = gateway
        with mock.patch.object(manager.LOG, 'warning') as logger_call, \
                mock.patch.object(socket.socket, 'connect') as mock_connect, \
                mock.patch.object(ovsdb_writer.OVSDBWriter,
                                  'delete_logical_switch') as mock_del_ls:
            mock_connect.side_effect = socket.error
            self.l2gw_agent_manager.delete_network(self.context,
                                                   mock.Mock(),
                                                   mock.Mock())
            self.assertEqual(1, logger_call.call_count)
            self.assertFalse(mock_del_ls.called)

    def test_init_with_enable_manager(self):
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        with mock.patch.object(manager.OVSDBManager,
                               '_sock_open_connection'
                               ) as mock_sock_open_conn, \
            mock.patch.object(loopingcall,
                              'FixedIntervalLoopingCall') as mock_loop:
            self.l2gw_agent_manager.__init__()
            self.assertTrue(mock_sock_open_conn.called)
            self.assertTrue(mock_loop.called)

    def test_sock_open_connection(self):
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common:
            self.l2gw_agent_manager.__init__()
            self.l2gw_agent_manager._sock_open_connection()
            self.assertTrue(mock_ovsdb_common.called)

    def test_set_monitor_agent_with_ovsdb_fd_None(self):
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.conf.host = 'fake_host'
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        self.l2gw_agent_manager.ovsdb_fd = None
        with mock.patch.object(manager.OVSDBManager,
                               '_sock_open_connection') as mock_open_conn:
            self.l2gw_agent_manager.set_monitor_agent(self.context,
                                                      'fake_host')
            self.assertTrue(mock_open_conn.called)

    def test_set_monitor_agent_with_ovsdb_fd_not_None(self):
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.conf.host = 'fake_host'
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common, \
            mock.patch.object(eventlet.greenthread,
                              'spawn_n') as mock_thread, \
            mock.patch.object(
                self.l2gw_agent_manager,
                '_start_looping_task_ovsdb_states') as mock_looping:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.ovsdb_fd.check_monitor_table_thread = False
            self.l2gw_agent_manager.ovsdb_fd.check_sock_rcv = True
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_conn_list = ["fake_ip"]
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_dicts = {
                "fake_ip": "fake_sock"}
            self.l2gw_agent_manager.set_monitor_agent(self.context,
                                                      'fake_host')
            self.assertTrue(mock_thread.called)
            self.assertTrue(mock_looping.called)

    def test_update_connection_to_gateway_for_monitor_agent(self):
        """Test case to test update_connection_to_gateway for

        monitor agent with enable_manager.
        """
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        fake_op_method = 'CREATE'
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.update_connection_to_gateway(
                self.context, mock.Mock(), mock.Mock(), mock.Mock(),
                mock.Mock(), mock.Mock(), fake_op_method)
            (self.l2gw_agent_manager.ovsdb_fd.update_connection_to_gateway.
             assert_called_with(mock.ANY, mock.ANY, mock.ANY, mock.ANY,
                                mock.ANY, fake_op_method, False))

    def test_update_connection_to_gateway_for_transact_agent(self):
        """Test case to test update_connection_to_gateway

        with enable_manager.
        """
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = ''
        fake_op_method = 'CREATE'
        with mock.patch.object(ovsdb_common_class, 'OVSDB_commom_class'
                               ) as mock_ovsdb_common, \
            mock.patch.object(manager.OVSDBManager,
                              '_sock_open_connection') as mock_open_conn:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_conn_list = ['fake_ip']
            self.l2gw_agent_manager.update_connection_to_gateway(
                self.context, 'fake_ip', mock.Mock(), mock.Mock(),
                mock.Mock(), mock.Mock(), fake_op_method)
            self.assertTrue(mock_open_conn.called)
            (self.l2gw_agent_manager.ovsdb_fd.update_connection_to_gateway.
             assert_called_with(mock.ANY, mock.ANY, mock.ANY, mock.ANY,
                                'fake_ip', fake_op_method, False))

    def test_update_connection_to_gateway_for_enable_manager_false(self):
        """Test case to test update_connection_to_gateway with

        enable_manager=False.
        """
        cfg.CONF.set_override('enable_manager', False, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        fake_op_method = 'CREATE'
        with mock.patch.object(self.l2gw_agent_manager,
                               '_is_valid_request', return_value=True
                               ) as mock_valid_req, \
            mock.patch.object(ovsdb_writer, 'OVSDBWriter'
                              ) as mock_ovsdb_fd:
            self.l2gw_agent_manager.update_connection_to_gateway(
                self.context, 'fake_ovsdb_id', "fake_logical_switch_dict",
                "fake_locator_dicts", "fake_mac_dicts", "fake_port_dicts",
                fake_op_method)
            ovsdb_sock_fd = mock_ovsdb_fd.return_value
            mock_valid_req.assert_called_with('fake_ovsdb_id')
            (ovsdb_sock_fd.update_connection_to_gateway.
             assert_called_with("fake_logical_switch_dict",
                                "fake_locator_dicts",
                                "fake_mac_dicts",
                                "fake_port_dicts",
                                "fake_ovsdb_id", fake_op_method))

    def test_delete_network_for_monitor_agent(self):
        """Test case to test delete_network with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.delete_network(
                self.context, mock.Mock(), "fake_logical_switch_uuid")
            (self.l2gw_agent_manager.ovsdb_fd.delete_logical_switch.
             assert_called_with("fake_logical_switch_uuid", mock.ANY, False))

    def test_delete_network_for_enable_manager_false(self):
        """Test case to test delete_network with enable_manager=False."""
        cfg.CONF.set_override('enable_manager', False, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        with mock.patch.object(self.l2gw_agent_manager,
                               '_is_valid_request', return_value=True
                               ) as mock_valid_req, \
            mock.patch.object(ovsdb_writer, 'OVSDBWriter'
                              ) as mock_ovsdb_fd:
            self.l2gw_agent_manager.delete_network(
                self.context, 'fake_ovsdb_id', "fake_logical_switch_uuid")
            ovsdb_sock_fd = mock_ovsdb_fd.return_value
            mock_valid_req.assert_called_with('fake_ovsdb_id')
            (ovsdb_sock_fd.delete_logical_switch.
             assert_called_with("fake_logical_switch_uuid", "fake_ovsdb_id"))

    def test_delete_network_for_transact_agent(self):
        """Test case to test delete_network with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = ''
        with mock.patch.object(ovsdb_common_class, 'OVSDB_commom_class'
                               ) as mock_ovsdb_common, \
            mock.patch.object(manager.OVSDBManager,
                              '_sock_open_connection') as mock_open_conn:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_conn_list = ['fake_ip']
            self.l2gw_agent_manager.delete_network(
                self.context, 'fake_ip', "fake_logical_switch_uuid")
            self.assertTrue(mock_open_conn.called)
            (self.l2gw_agent_manager.ovsdb_fd.delete_logical_switch.
             assert_called_with("fake_logical_switch_uuid", 'fake_ip', False))

    def test_add_vif_to_gateway_for_monitor_agent(self):
        """Test case to test add_vif_to_gateway with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.add_vif_to_gateway(
                self.context, mock.Mock(), "fake_logical_switch_dict",
                "fake_locator_dict", "fake_mac_dict")
            (self.l2gw_agent_manager.ovsdb_fd.insert_ucast_macs_remote.
             assert_called_with("fake_logical_switch_dict",
                                "fake_locator_dict", "fake_mac_dict",
                                mock.ANY, False))

    def test_add_vif_to_gateway_for_transact_agent(self):
        """Test case to test add_vif_to_gateway with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = ''
        with mock.patch.object(ovsdb_common_class, 'OVSDB_commom_class'
                               ) as mock_ovsdb_common, \
            mock.patch.object(manager.OVSDBManager,
                              '_sock_open_connection') as mock_open_conn:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.ovsdb_fd.check_c_sock = True
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_conn_list = ['fake_ip']
            self.l2gw_agent_manager.add_vif_to_gateway(
                self.context, 'fake_ip', "fake_logical_switch_dict",
                "fake_locator_dict", "fake_mac_dict")
            self.assertTrue(mock_open_conn.called)
            (self.l2gw_agent_manager.ovsdb_fd.insert_ucast_macs_remote.
             assert_called_with("fake_logical_switch_dict",
                                "fake_locator_dict", "fake_mac_dict",
                                'fake_ip', False))

    def test_add_vif_to_gateway_for_enable_manager_false(self):
        """Test case to test add_vif_to_gateway with

        enable_manager=False.
        """
        cfg.CONF.set_override('enable_manager', False, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        with mock.patch.object(self.l2gw_agent_manager,
                               '_is_valid_request',
                               return_value=True) as mock_valid_req, \
                mock.patch.object(ovsdb_writer, 'OVSDBWriter'
                                  ) as mock_ovsdb_fd:
            self.l2gw_agent_manager.add_vif_to_gateway(
                self.context, 'fake_ovsdb_id', "fake_logical_switch_dict",
                "fake_locator_dict", "fake_mac_dict")
            ovsdb_sock_fd = mock_ovsdb_fd.return_value
            mock_valid_req.assert_called_with('fake_ovsdb_id')
            (ovsdb_sock_fd.insert_ucast_macs_remote.
             assert_called_with("fake_logical_switch_dict",
                                "fake_locator_dict",
                                "fake_mac_dict",
                                "fake_ovsdb_id"))

    def test_delete_vif_from_gateway_for_monitor_agent(self):
        """Test case to test delete_vif_to_gateway with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.delete_vif_from_gateway(
                self.context, mock.Mock(), "fake_logical_switch_uuid",
                "fake_mac")
            (self.l2gw_agent_manager.ovsdb_fd.delete_ucast_macs_remote.
             assert_called_with("fake_logical_switch_uuid", "fake_mac",
                                mock.ANY, False))

    def test_delete_vif_from_gateway_for_transact_agent(self):
        """Test case to test delete_vif_from_gateway with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = ''
        with mock.patch.object(ovsdb_common_class, 'OVSDB_commom_class'
                               ) as mock_ovsdb_common, \
            mock.patch.object(manager.OVSDBManager,
                              '_sock_open_connection') as mock_open_conn:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_conn_list = ['fake_ip']
            self.l2gw_agent_manager.delete_vif_from_gateway(
                self.context, 'fake_ip', "fake_logical_switch_uuid",
                "fake_mac")
            self.assertTrue(mock_open_conn.called)
            (self.l2gw_agent_manager.ovsdb_fd.delete_ucast_macs_remote.
             assert_called_with("fake_logical_switch_uuid", "fake_mac",
                                'fake_ip', False))

    def test_delete_vif_from_gateway_for_enable_manager_false(self):
        """Test case to test delete_vif_from_gateway with

        enable_manager=False.
        """
        cfg.CONF.set_override('enable_manager', False, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        with mock.patch.object(self.l2gw_agent_manager,
                               '_is_valid_request', return_value=True
                               ) as mock_valid_req, \
            mock.patch.object(ovsdb_writer, 'OVSDBWriter'
                              ) as mock_ovsdb_fd:
            self.l2gw_agent_manager.delete_vif_from_gateway(
                self.context, 'fake_ovsdb_id', "fake_logical_switch_uuid",
                "fake_mac")
            ovsdb_sock_fd = mock_ovsdb_fd.return_value
            mock_valid_req.assert_called_with('fake_ovsdb_id')
            (ovsdb_sock_fd.delete_ucast_macs_remote.
             assert_called_with("fake_logical_switch_uuid",
                                "fake_mac",
                                "fake_ovsdb_id"))

    def test_update_vif_to_gateway_for_monitor_agent(self):
        """Test case to test update_vif_to_gateway with enable_manager."""
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = n_const.MONITOR
        with mock.patch.object(ovsdb_common_class,
                               'OVSDB_commom_class') as mock_ovsdb_common:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.update_vif_to_gateway(
                self.context, mock.Mock(),
                "fake_logical_switch_uuid", "fake_mac")
            (self.l2gw_agent_manager.ovsdb_fd.update_ucast_macs_remote.
             assert_called_with(
                 "fake_logical_switch_uuid", "fake_mac", mock.ANY, False))

    def test_update_vif_to_gateway_for_transact_agent(self):
        """Test case to test update_vif_to_gateway

           with enable_manager.
        """
        cfg.CONF.set_override('enable_manager', True, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        self.l2gw_agent_manager.l2gw_agent_type = ''
        with mock.patch.object(ovsdb_common_class, 'OVSDB_commom_class'
                               ) as mock_ovsdb_common, \
            mock.patch.object(manager.OVSDBManager,
                              '_sock_open_connection') as mock_open_conn:
            self.l2gw_agent_manager.ovsdb_fd = mock_ovsdb_common.return_value
            self.l2gw_agent_manager.ovsdb_fd.ovsdb_conn_list = ['fake_ip']
            self.l2gw_agent_manager.update_vif_to_gateway(
                self.context, 'fake_ip', "fake_logical_switch_uuid",
                "fake_mac")
            self.assertTrue(mock_open_conn.called)
            (self.l2gw_agent_manager.ovsdb_fd.update_ucast_macs_remote.
             assert_called_with("fake_logical_switch_uuid", "fake_mac",
                                'fake_ip', False))

    def test_update_vif_to_gateway_for_enable_manager_false(self):
        """Test case to test update_vif_to_gateway with

        enable_manager=False.
        """
        cfg.CONF.set_override('enable_manager', False, 'ovsdb')
        self.l2gw_agent_manager.__init__()
        with mock.patch.object(self.l2gw_agent_manager,
                               '_is_valid_request', return_value=True
                               ) as mock_valid_req, \
            mock.patch.object(ovsdb_writer, 'OVSDBWriter'
                              ) as mock_ovsdb_fd:
            self.l2gw_agent_manager.update_vif_to_gateway(
                self.context, 'fake_ovsdb_id', "fake_locator_dict",
                "fake_mac_dict")
            ovsdb_sock_fd = mock_ovsdb_fd.return_value
            mock_valid_req.assert_called_with('fake_ovsdb_id')
            (ovsdb_sock_fd.update_ucast_macs_remote.
             assert_called_with("fake_locator_dict",
                                "fake_mac_dict",
                                "fake_ovsdb_id"))



# Copyright 2015 OpenStack Foundation
# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from neutron_lib.api import extensions as api_extensions
from neutron_lib.api import validators

from neutron.api import extensions
from neutron.api.v2 import resource_helper
from neutron.common import config as common_config

from networking_l2gw import extensions as l2gw_extensions
from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway.common import l2gw_validators

common_config.register_common_config_options()
extensions.append_api_extensions_path(l2gw_extensions.__path__)

RESOURCE_ATTRIBUTE_MAP = {
    constants.L2_GATEWAYS: {
        'id': {'allow_post': False, 'allow_put': False,
               'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'is_visible': True, 'default': ''},
        'devices': {'allow_post': True, 'allow_put': True,
                    'validate': {'type:l2gwdevice_list': None},
                    'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True,
                      'is_visible': True}
    },
}

validators.add_validator('l2gwdevice_list',
                         l2gw_validators.validate_gwdevice_list)


class L2gateway(api_extensions.ExtensionDescriptor):

    """API extension for Layer-2 Gateway support."""

    @classmethod
    def get_name(cls):
        return "L2 Gateway"

    @classmethod
    def get_alias(cls):
        return "l2-gateway"

    @classmethod
    def get_description(cls):
        return "Connects Neutron networks with external networks at layer 2."

    @classmethod
    def get_updated(cls):
        return "2015-01-01T00:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        resources = resource_helper.build_resource_info(plural_mappings,
                                                        RESOURCE_ATTRIBUTE_MAP,
                                                        constants.L2GW)

        return resources

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


class L2GatewayPluginBase(object):

    @abc.abstractmethod
    def create_l2_gateway(self, context, l2_gateway):
        pass

    @abc.abstractmethod
    def get_l2_gateway(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def delete_l2_gateway(self, context, id):
        pass

    @abc.abstractmethod
    def get_l2_gateways(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        pass

    @abc.abstractmethod
    def update_l2_gateway(self, context, id, l2_gateway):
        pass



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from neutron.agent.common import utils
from neutron.db import agents_db
from neutron_lib import context as neutron_context

from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway.common import constants as srv_const

LOG = logging.getLogger(__name__)
# DNM: Do Not Merge, test patch only!


class L2GatewayAgentScheduler(agents_db.AgentDbMixin):
    """L2gateway agent scheduler class.

    This maintains active and inactive agents and
    select monitor and transact agents.
    """
    _plugin = None
    _l2gwplugin = None

    def __init__(self, agent_rpc, notifier=None):
        super(L2GatewayAgentScheduler, self).__init__()
        self.notifier = notifier
        config.register_l2gw_opts_helper()
        self.monitor_interval = cfg.CONF.periodic_monitoring_interval
        self.agent_rpc = agent_rpc

    @property
    def l2gwplugin(self):
        if self._l2gwplugin is None:
            self._l2gwplugin = directory.get_plugin(srv_const.L2GW)
        return self._l2gwplugin

    @property
    def plugin(self):
        if self._plugin is None:
            self._plugin = directory.get_plugin()
        return self._plugin

    def initialize_thread(self):
        """Initialization of L2gateway agent scheduler thread."""
        try:
            monitor_thread = loopingcall.FixedIntervalLoopingCall(
                self.monitor_agent_state)
            monitor_thread.start(
                interval=self.monitor_interval,
                initial_delay=random.randint(self.monitor_interval,
                                             self.monitor_interval * 2))
            LOG.debug("Successfully initialized L2gateway agent scheduler"
                      " thread with loop interval %s", self.monitor_interval)
        except Exception:
            LOG.error("Cannot initialize agent scheduler thread")

    def _select_agent_type(self, context, agents_to_process):
        """Select the Monitor agent."""
        # Various cases to be handled:
        # 1. Check if there is a single active L2 gateway agent.
        #    If only one agent is active, then make it the Monitor agent.
        # 2. Else, in the list of the active agents, if there does not
        #    exist Monitor agent, then make the agent that
        #    started first as the Monitor agent.
        # 3. If multiple Monitor agents exist (case where the Monitor agent
        #    gets disconnected from the Neutron server and another agent
        #    becomes the Monitor agent and then the original Monitor agent
        #    connects back within the agent downtime value), then we need to
        #    send the fanout message so that only one becomes the Monitor
        #    agent.

        # Check if there already exists Monitor agent and it's the only one.
        monitor_agents = [
            x for x in agents_to_process
            if x['configurations'].get(srv_const.L2GW_AGENT_TYPE) ==
            srv_const.MONITOR]
        if len(monitor_agents) == 1:
            return

        # We either have more than one Monitor agent,
        # or there does not exist Monitor agent.
        # We will decide which agent should be the Monitor agent.
        chosen_agent = None
        if len(agents_to_process) == 1:
            # Only one agent is configured.
            # Make it the Monitor agent
            chosen_agent = agents_to_process[0]
        else:
            # Select the agent with the oldest started_at
            # timestamp as the Monitor agent.
            sorted_active_agents = sorted(agents_to_process,
                                          key=lambda k: k['started_at'])
            chosen_agent = sorted_active_agents[0]
        self.agent_rpc.set_monitor_agent(context, chosen_agent['host'])

    def monitor_agent_state(self):
        """Represents L2gateway agent scheduler thread.

        Maintains list of active and inactive agents based on
        the heartbeat recorded.
        """
        context = neutron_context.get_admin_context()
        try:
            all_agents = self.plugin.get_agents(
                context,
                filters={'agent_type': [srv_const.AGENT_TYPE_L2GATEWAY]})
        except Exception:
            LOG.exception("Unable to get the agent list. Continuing...")
            return

        # Reset the agents that will be processed for selecting the
        # Monitor agent
        agents_to_process = []
        for agent in all_agents:
            if not utils.is_agent_down(agent['heartbeat_timestamp']):
                agents_to_process.append(agent)
        if agents_to_process:
            self._select_agent_type(context, agents_to_process)
        return



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils

from networking_l2gw.services.l2gateway.agent.ovsdb import base_connection
from networking_l2gw.services.l2gateway.common import constants as n_const
from networking_l2gw.services.l2gateway.common import ovsdb_schema
from networking_l2gw.services.l2gateway import exceptions

LOG = logging.getLogger(__name__)


class Activity(object):
    Initial, Update = range(2)


class OVSDBMonitor(base_connection.BaseConnection):
    """Monitors OVSDB servers."""
    def __init__(self, conf, gw_config, callback, mgr=None):
        super(OVSDBMonitor, self).__init__(conf, gw_config, mgr=None)
        self.mgr = mgr
        self.rpc_callback = callback
        self.callbacks = {}
        self._setup_dispatch_table()
        self.read_on = True
        self.handlers = {"echo": self._default_echo_handler}
        self.sock_timeout = cfg.CONF.ovsdb.socket_timeout
        if self.enable_manager:
            self.check_monitor_table_thread = False
        if not self.enable_manager:
            eventlet.greenthread.spawn(self._rcv_thread)

    def _spawn_monitor_table_thread(self, addr):
        self.set_monitor_response_handler(addr)
        self.check_monitor_table_thread = True

    def _initialize_data_dict(self):
        data_dict = {'new_local_macs': [],
                     'deleted_local_macs': [],
                     'modified_local_macs': [],
                     'new_remote_macs': [],
                     'deleted_remote_macs': [],
                     'modified_remote_macs': [],
                     'new_physical_ports': [],
                     'deleted_physical_ports': [],
                     'modified_physical_ports': [],
                     'new_physical_switches': [],
                     'deleted_physical_switches': [],
                     'modified_physical_switches': [],
                     'new_physical_locators': [],
                     'deleted_physical_locators': [],
                     'modified_physical_locators': [],
                     'new_logical_switches': [],
                     'deleted_logical_switches': [],
                     'modified_logical_switches': [],
                     'new_mlocal_macs': [],
                     'deleted_mlocal_macs': [],
                     'modified_mlocal_macs': [],
                     'new_locator_sets': [],
                     'deleted_locator_sets': [],
                     'modified_locator_sets': []}
        return data_dict

    def _setup_dispatch_table(self):
        self.dispatch_table = {'Logical_Switch': self._process_logical_switch,
                               'Ucast_Macs_Local':
                               self._process_ucast_macs_local,
                               'Physical_Locator':
                               self._process_physical_locator,
                               'Ucast_Macs_Remote':
                               self._process_ucast_macs_remote,
                               'Mcast_Macs_Local':
                               self._process_mcast_macs_local,
                               'Physical_Locator_Set':
                               self._process_physical_locator_set
                               }

    def set_monitor_response_handler(self, addr=None):
        """Monitor OVSDB tables to receive events for any changes in OVSDB."""
        if self.connected:
            op_id = str(random.getrandbits(128))
            props = {'select': {'initial': True,
                                'insert': True,
                                'delete': True,
                                'modify': True}}
            monitor_message = {'id': op_id,
                               'method': 'monitor',
                               'params': [n_const.OVSDB_SCHEMA_NAME,
                                          None,
                                          {'Logical_Switch': [props],
                                           'Physical_Switch': [props],
                                           'Physical_Port': [props],
                                           'Ucast_Macs_Local': [props],
                                           'Ucast_Macs_Remote': [props],
                                           'Physical_Locator': [props],
                                           'Mcast_Macs_Local': [props],
                                           'Physical_Locator_Set': [props]}
                                          ]}
            self._set_handler("update", self._update_event_handler)
            if not self.send(monitor_message, addr=addr):
                # Return so that this will retried in the next iteration
                return
            try:
                response_result = self._process_response(op_id)
            except exceptions.OVSDBError:
                with excutils.save_and_reraise_exception():
                    if self.enable_manager:
                        self.check_monitor_table_thread = False
                    LOG.exception("Exception while receiving the "
                                  "response for the monitor message")
            self._process_monitor_msg(response_result, addr)

    def _update_event_handler(self, message, addr):
        self._process_update_event(message, addr)

    def _process_update_event(self, message, addr):
        """Process update event that is triggered by the OVSDB server."""
        LOG.debug("_process_update_event: message = %s ", str(message))
        data_dict = self._initialize_data_dict()
        if message.get('method') == 'update':
            params_list = message.get('params')
            param_dict = params_list[1]
            self._process_tables(param_dict, data_dict)
            self.rpc_callback(Activity.Update,
                              self._form_ovsdb_data(data_dict, addr))

    def _process_tables(self, param_dict, data_dict):
        # Process all the tables one by one.
        # OVSDB table name is the key in the dictionary.
        port_map = {}
        for table_name in param_dict.keys():
            table_dict = param_dict.get(table_name)
            for uuid in table_dict.keys():
                uuid_dict = table_dict.get(uuid)
                if table_name == 'Physical_Switch':
                    self._process_physical_switch(uuid,
                                                  uuid_dict,
                                                  port_map,
                                                  data_dict)
                elif table_name == 'Physical_Port':
                    self._process_physical_port(uuid, uuid_dict,
                                                port_map, data_dict)
                else:
                    self.dispatch_table.get(table_name)(uuid, uuid_dict,
                                                        data_dict)

    def _process_response(self, op_id):
        result = self._response(op_id)
        count = 0
        while (not result and count <= n_const.MAX_RETRIES):
            count = count + 1
            eventlet.greenthread.sleep(0)
            result = self._response(op_id)
        if not result and count >= n_const.MAX_RETRIES:
            raise exceptions.OVSDBError(
                message="OVSDB server did not respond within "
                "max retry attempts.")
        error = result.get("error", None)
        if error:
            raise exceptions.OVSDBError(
                message="Error from the OVSDB server %s" % error
            )
        return result

    def _default_echo_handler(self, message, addr):
        """Message handler for the OVSDB server's echo request."""
        self.send({"result": message.get("params", None),
                   "error": None, "id": message['id']}, addr=addr)

    def _set_handler(self, method_name, handler):
        self.handlers[method_name] = handler

    def _on_remote_message(self, message, addr=None):
        """Processes the message received on the socket."""
        try:
            json_m = jsonutils.loads(message)
            handler_method = json_m.get('method', None)
            if handler_method:
                self.handlers.get(handler_method)(json_m, addr)
            else:
                self.responses.append(json_m)
        except Exception as e:
            LOG.exception("Exception [%s] while handling "
                          "message", e)

    def _rcv_thread(self):
        chunks = []
        lc = rc = 0
        prev_char = None
        while self.read_on:
            try:
                # self.socket.recv() is a blocked call
                # (if timeout value is not passed) due to which we cannot
                # determine if the remote OVSDB server has died. The remote
                # OVSDB server sends echo requests every 4 seconds.
                # If there is no echo request on the socket for socket_timeout
                # seconds(by default its 30 seconds),
                # the agent can safely assume that the connection with the
                # remote OVSDB server is lost. Better to retry by reopening
                # the socket.
                self.socket.settimeout(self.sock_timeout)
                response = self.socket.recv(n_const.BUFFER_SIZE)
                eventlet.greenthread.sleep(0)
                if response:
                    response = response.decode('utf8')
                    message_mark = 0
                    for i, c in enumerate(response):
                        if c == '{' and not (prev_char and
                                             prev_char == '\\'):
                            lc += 1
                        elif c == '}' and not (prev_char and
                                               prev_char == '\\'):
                            rc += 1
                        if rc > lc:
                            raise Exception("json string not valid")
                        elif lc == rc and lc != 0:
                            chunks.append(response[message_mark:i + 1])
                            message = "".join(chunks)
                            eventlet.greenthread.spawn_n(
                                self._on_remote_message, message)
                            lc = rc = 0
                            message_mark = i + 1
                            chunks = []
                        prev_char = c
                    chunks.append(response[message_mark:])
                else:
                    self.read_on = False
                    self.disconnect()
            except Exception as ex:
                self.read_on = False
                self.disconnect()
                LOG.exception("Exception [%s] occurred while receiving"
                              "message from the OVSDB server", ex)

    def disconnect(self, addr=None):
        """disconnects the connection from the OVSDB server."""
        self.read_on = False
        super(OVSDBMonitor, self).disconnect(addr)

    def _process_monitor_msg(self, message, addr=None):
        """Process initial set of records in the OVSDB at startup."""
        result_dict = message.get('result')
        data_dict = self._initialize_data_dict()
        try:
            self._process_tables(result_dict, data_dict)
            self.rpc_callback(Activity.Initial,
                              self._form_ovsdb_data(data_dict, addr))
        except Exception as e:
            LOG.exception("_process_monitor_msg:ERROR %s ", e)

    def _get_list(self, resource_list):
        return [element.__dict__ for element in resource_list]

    def _form_ovsdb_data(self, data_dict, addr):
        return {n_const.OVSDB_IDENTIFIER: str(addr) if (
                self.enable_manager) else (self.gw_config.ovsdb_identifier),
                'new_logical_switches': self._get_list(
                    data_dict.get('new_logical_switches')),
                'new_physical_switches': self._get_list(
                    data_dict.get('new_physical_switches')),
                'new_physical_ports': self._get_list(
                    data_dict.get('new_physical_ports')),
                'new_physical_locators': self._get_list(
                    data_dict.get('new_physical_locators')),
                'new_local_macs': self._get_list(
                    data_dict.get('new_local_macs')),
                'new_remote_macs': self._get_list(
                    data_dict.get('new_remote_macs')),
                'new_mlocal_macs': self._get_list(
                    data_dict.get('new_mlocal_macs')),
                'new_locator_sets': self._get_list(
                    data_dict.get('new_locator_sets')),
                'deleted_logical_switches': self._get_list(
                    data_dict.get('deleted_logical_switches')),
                'deleted_physical_switches': self._get_list(
                    data_dict.get('deleted_physical_switches')),
                'deleted_physical_ports': self._get_list(
                    data_dict.get('deleted_physical_ports')),
                'deleted_physical_locators': self._get_list(
                    data_dict.get('deleted_physical_locators')),
                'deleted_local_macs': self._get_list(
                    data_dict.get('deleted_local_macs')),
                'deleted_remote_macs': self._get_list(
                    data_dict.get('deleted_remote_macs')),
                'deleted_mlocal_macs': self._get_list(
                    data_dict.get('deleted_mlocal_macs')),
                'deleted_locator_sets': self._get_list(
                    data_dict.get('deleted_locator_sets')),
                'modified_logical_switches': self._get_list(
                    data_dict.get('modified_logical_switches')),
                'modified_physical_switches': self._get_list(
                    data_dict.get('modified_physical_switches')),
                'modified_physical_ports': self._get_list(
                    data_dict.get('modified_physical_ports')),
                'modified_physical_locators': self._get_list(
                    data_dict.get('modified_physical_locators')),
                'modified_local_macs': self._get_list(
                    data_dict.get('modified_local_macs')),
                'modified_remote_macs': self._get_list(
                    data_dict.get('modified_remote_macs')),
                'modified_mlocal_macs': self._get_list(
                    data_dict.get('modified_mlocal_macs')),
                'modified_locator_sets': self._get_list(
                    data_dict.get('modified_locator_sets'))}

    def _process_physical_port(self, uuid, uuid_dict, port_map, data_dict):
        """Processes Physical_Port record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            port_fault_status = new_row.get('port_fault_status')
            if type(port_fault_status) is list:
                port_fault_status = None
            port = ovsdb_schema.PhysicalPort(uuid, new_row.get('name'), None,
                                             None,
                                             port_fault_status)
            switch_id = port_map.get(uuid, None)
            if switch_id:
                port.physical_switch_id = switch_id
            # Update the vlan bindings
            outer_binding_list = new_row.get('vlan_bindings')
            # First element is "map"
            outer_binding_list.remove(outer_binding_list[0])
            vlan_bindings = []
            if len(outer_binding_list) > 0:
                for binding in outer_binding_list:
                    if len(binding) > 0:
                        for element in binding:
                            vlan = element[0]
                            inner_most_list = element[1]
                            ls_id = inner_most_list[1]
                            vb = ovsdb_schema.VlanBinding(vlan, ls_id).__dict__
                            vlan_bindings.append(vb)
            port.vlan_bindings = vlan_bindings
            if old_row:
                modified_physical_ports = data_dict.get(
                    'modified_physical_ports')
                modified_physical_ports.append(port)
            else:
                new_physical_ports = data_dict.get(
                    'new_physical_ports')
                new_physical_ports.append(port)
        elif old_row:
            # Port is deleted permanently from OVSDB server
            port_fault_status = old_row.get('port_fault_status')
            if type(port_fault_status) is list:
                port_fault_status = None
            port = ovsdb_schema.PhysicalPort(uuid, old_row.get('name'), None,
                                             None,
                                             port_fault_status)
            deleted_physical_ports = data_dict.get('deleted_physical_ports')
            deleted_physical_ports.append(port)

    def _process_physical_switch(self, uuid, uuid_dict, port_map, data_dict):
        """Processes Physical_Switch record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            # insert or modify operation
            ports = new_row.get('ports')
            # First element in the list is either 'set' or 'uuid'
            # Let us remove it.
            is_set = False
            if ports[0] == 'set':
                is_set = True
            ports.remove(ports[0])
            all_ports = []
            if not is_set:
                all_ports.append(ports[0])
            else:
                for port_list in ports:
                    # each port variable is again list
                    for port in port_list:
                        for inner_port in port:
                            if inner_port != 'uuid':
                                all_ports.append(inner_port)
            switch_fault_status = new_row.get('switch_fault_status')
            if type(switch_fault_status) is list:
                switch_fault_status = None
            phys_switch = ovsdb_schema.PhysicalSwitch(
                uuid, new_row.get('name'), new_row.get('tunnel_ips'),
                switch_fault_status)
            # Now, store mapping of physical ports to
            # physical switch so that it is useful while
            # processing Physical_Switch record
            for port in all_ports:
                port_map[port] = uuid
                for pport in data_dict['new_physical_ports']:
                    if pport.uuid == port:
                        pport.physical_switch_id = uuid
            if old_row:
                modified_physical_switches = data_dict.get(
                    'modified_physical_switches')
                modified_physical_switches.append(phys_switch)
            else:
                new_physical_switches = data_dict.get(
                    'new_physical_switches')
                new_physical_switches.append(phys_switch)
        elif old_row:
            # Physical switch is deleted permanently from OVSDB
            # server
            switch_fault_status = old_row.get('switch_fault_status')
            if type(switch_fault_status) is list:
                switch_fault_status = None
            phys_switch = ovsdb_schema.PhysicalSwitch(
                uuid, old_row.get('name'), old_row.get('tunnel_ips'),
                switch_fault_status)
            deleted_physical_switches = data_dict.get(
                'deleted_physical_switches')
            deleted_physical_switches.append(phys_switch)

    def _process_logical_switch(self, uuid, uuid_dict, data_dict):
        """Processes Logical_Switch record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            l_switch = ovsdb_schema.LogicalSwitch(uuid,
                                                  new_row.get('name'),
                                                  new_row.get('tunnel_key'),
                                                  new_row.get('description'))
            if old_row:
                modified_logical_switches = data_dict.get(
                    'modified_logical_switches')
                modified_logical_switches.append(l_switch)
            else:
                new_logical_switches = data_dict.get(
                    'new_logical_switches')
                new_logical_switches.append(l_switch)
        elif old_row:
            l_switch = ovsdb_schema.LogicalSwitch(uuid,
                                                  old_row.get('name'),
                                                  old_row.get('tunnel_key'),
                                                  old_row.get('description'))
            deleted_logical_switches = data_dict.get(
                'deleted_logical_switches')
            deleted_logical_switches.append(l_switch)

    def _process_ucast_macs_local(self, uuid, uuid_dict, data_dict):
        """Processes Ucast_Macs_Local record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            locator_list = new_row.get('locator')
            locator_id = locator_list[1]
            logical_switch_list = new_row.get('logical_switch')
            logical_switch_id = logical_switch_list[1]
            mac_local = ovsdb_schema.UcastMacsLocal(uuid,
                                                    new_row.get('MAC'),
                                                    logical_switch_id,
                                                    locator_id,
                                                    new_row.get('ipaddr'))
            if old_row:
                modified_local_macs = data_dict.get(
                    'modified_local_macs')
                modified_local_macs.append(mac_local)
            else:
                new_local_macs = data_dict.get(
                    'new_local_macs')
                new_local_macs.append(mac_local)
        elif old_row:
            # A row from UcastMacLocal is deleted.
            logical_switch_list = old_row.get('logical_switch')
            l_sw_id = logical_switch_list[1]
            mac_local = ovsdb_schema.UcastMacsLocal(uuid,
                                                    old_row.get('MAC'),
                                                    l_sw_id,
                                                    None,
                                                    None)
            deleted_local_macs = data_dict.get(
                'deleted_local_macs')
            deleted_local_macs.append(mac_local)

    def _process_ucast_macs_remote(self, uuid, uuid_dict, data_dict):
        """Processes Ucast_Macs_Remote record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            locator_list = new_row.get('locator')
            locator_id = locator_list[1]
            logical_switch_list = new_row.get('logical_switch')
            logical_switch_id = logical_switch_list[1]
            mac_remote = ovsdb_schema.UcastMacsRemote(uuid,
                                                      new_row.get('MAC'),
                                                      logical_switch_id,
                                                      locator_id,
                                                      new_row.get('ipaddr'))
            if old_row:
                modified_remote_macs = data_dict.get(
                    'modified_remote_macs')
                modified_remote_macs.append(mac_remote)
            else:
                new_remote_macs = data_dict.get(
                    'new_remote_macs')
                new_remote_macs.append(mac_remote)
        elif old_row:
            logical_switch_list = old_row.get('logical_switch')
            l_sw_id = logical_switch_list[1]
            mac_remote = ovsdb_schema.UcastMacsRemote(uuid,
                                                      old_row.get('MAC'),
                                                      l_sw_id,
                                                      None,
                                                      None)
            deleted_remote_macs = data_dict.get(
                'deleted_remote_macs')
            deleted_remote_macs.append(mac_remote)

    def _process_physical_locator(self, uuid, uuid_dict, data_dict):
        """Processes Physical_Locator record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            dstip = new_row['dst_ip']
            locator = ovsdb_schema.PhysicalLocator(uuid, dstip)
            if old_row:
                modified_physical_locators = data_dict.get(
                    'modified_physical_locators')
                modified_physical_locators.append(locator)
            else:
                new_physical_locators = data_dict.get(
                    'new_physical_locators')
                new_physical_locators.append(locator)
        elif old_row:
            dstip = old_row['dst_ip']
            locator = ovsdb_schema.PhysicalLocator(uuid, dstip)
            deleted_physical_locators = data_dict.get(
                'deleted_physical_locators')
            deleted_physical_locators.append(locator)

    def _process_mcast_macs_local(self, uuid, uuid_dict, data_dict):
        """Processes Mcast_Macs_Local record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            locator_set_list = new_row.get('locator_set')
            logical_switch_list = new_row.get('logical_switch')
            mcast_local = ovsdb_schema.McastMacsLocal(uuid,
                                                      new_row['MAC'],
                                                      logical_switch_list[1],
                                                      locator_set_list[1],
                                                      new_row['ipaddr'])
            if old_row:
                modified_mlocal_macs = data_dict.get(
                    'modified_mlocal_macs')
                modified_mlocal_macs.append(mcast_local)
            else:
                new_mlocal_macs = data_dict.get(
                    'new_mlocal_macs')
                new_mlocal_macs.append(mcast_local)
        elif old_row:
            logical_switch_list = old_row.get('logical_switch')
            l_sw_id = logical_switch_list[1]
            mcast_local = ovsdb_schema.McastMacsLocal(uuid,
                                                      old_row.get('MAC'),
                                                      l_sw_id,
                                                      None,
                                                      None)
            deleted_mlocal_macs = data_dict.get(
                'deleted_mlocal_macs')
            deleted_mlocal_macs.append(mcast_local)

    def _process_physical_locator_set(self, uuid, uuid_dict, data_dict):
        """Processes Physical_Locator_Set record from the OVSDB event."""
        new_row = uuid_dict.get('new', None)
        old_row = uuid_dict.get('old', None)
        if new_row:
            locator_set = self._form_locator_set(uuid, new_row)
            if old_row:
                modified_locator_sets = data_dict.get(
                    'modified_locator_sets')
                modified_locator_sets.append(locator_set)
            else:
                new_locator_sets = data_dict.get(
                    'new_locator_sets')
                new_locator_sets.append(locator_set)
        elif old_row:
            locator_set = self._form_locator_set(uuid, old_row)
            deleted_locator_sets = data_dict.get(
                'deleted_locator_sets')
            deleted_locator_sets.append(locator_set)

    def _form_locator_set(self, uuid, row):
        locators = []
        locator_set_list = row.get('locators')
        if locator_set_list[0] == 'set':
            locator_set_list = locator_set_list[1]
            for locator in locator_set_list:
                locators.append(locator[1])
        else:
            locators.append(locator_set_list[1])
        locator_set = ovsdb_schema.PhysicalLocatorSet(uuid,
                                                      locators)
        return locator_set



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from networking_l2gw.services.l2gateway.common import constants as n_const

OVSDB_IP = 'ovsdb_ip'
OVSDB_PORT = 'ovsdb_port'
PRIVATE_KEY = 'private_key'
USE_SSL = 'use_ssl'
CERTIFICATE = 'certificate'
CA_CERT = 'ca_cert'


class L2GatewayConfig(object):
    def __init__(self, ovsdb_config):
        self.use_ssl = False
        if ovsdb_config.get(USE_SSL, None):
            self.use_ssl = ovsdb_config[USE_SSL]
            self.private_key = ovsdb_config[PRIVATE_KEY]
            self.certificate = ovsdb_config[CERTIFICATE]
            self.ca_cert = ovsdb_config[CA_CERT]

        self.ovsdb_identifier = ovsdb_config[n_const.OVSDB_IDENTIFIER]
        self.ovsdb_ip = ovsdb_config[OVSDB_IP]
        self.ovsdb_port = ovsdb_config[OVSDB_PORT]
        self.ovsdb_fd = None



# Copyright (c) 2016 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from ovs.db import idl

from ovsdbapp.backend.ovs_idl import connection as conn


def get_schema_helper_for_vtep():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    return idl.SchemaHelper(current_dir + '/../vtep/vtep.ovsschema')


class Connection(conn.Connection):
    def __init__(self, connection, timeout, schema_name):
        idl_ = idl.Idl(connection, get_schema_helper_for_vtep())
        super(Connection, self).__init__(idl_, timeout)



# Copyright (c) 2016 Openstack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

from neutron.tests import base

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway import plugin as l2gw_plugin


class TestL2GatewayPlugin(base.BaseTestCase):

    def setUp(self):
        super(TestL2GatewayPlugin, self).setUp()
        mock.patch.object(config, 'register_l2gw_opts_helper')
        self.driver = mock.MagicMock()
        mock.patch('neutron.services.service_base.load_drivers',
                   return_value=({'dummyprovider': self.driver},
                                 'dummyprovider')).start()
        mock.patch.object(l2gateway_db.L2GatewayMixin, '__init__'),
        mock.patch.object(l2gateway_db, 'subscribe')
        mock.patch('neutron.db.servicetype_db.ServiceTypeManager.get_instance',
                   return_value=mock.MagicMock()).start()
        self.context = mock.MagicMock()
        self.plugin = l2gw_plugin.L2GatewayPlugin()
        self.ovsdb_identifier = 'fake_ovsdb_id'

    def _get_fake_l2_gateway(self):
        fake_l2_gateway_id = "5227c228-6bba-4bbe-bdb8-6942768ff0f1"
        fake_l2_gateway = {
            "tenant_id": "de0a7495-05c4-4be0-b796-1412835c6820",
            "id": "5227c228-6bba-4bbe-bdb8-6942768ff0f1",
            "name": "test-gateway",
            "devices": [
                {
                    "device_name": "switch1",
                    "interfaces": [
                        {
                            "name": "port1",
                            "segmentation_id": [100]
                        },
                        {
                            "name": "port2",
                            "segmentation_id": [151, 152]
                        }
                    ]
                }
            ]
        }
        return fake_l2_gateway_id, fake_l2_gateway

    def _get_fake_l2_gateway_connection(self):
        fake_l2_gateway_conn_id = "5227c228-6bba-4bbe-bdb8-6942768ff02f"
        fake_l2_gateway_conn = {
            "tenant_id": "de0a7495-05c4-4be0-b796-1412835c6820",
            "id": "5227c228-6bba-4bbe-bdb8-6942768ff02f",
            "default_segmentation_id": 77,
            "network_id": "5227c228-6bba-4bbe-bdb8-6942768ff077",
            "l2_gateway_id": "4227c228-6bba-4bbe-bdb8-6942768ff088"
        }
        return fake_l2_gateway_conn_id, fake_l2_gateway_conn

    def test_add_port_mac(self):
        self.plugin.add_port_mac(self.context, {})
        self.driver.add_port_mac.assert_called_once_with(self.context, {})

    def test_delete_port_mac(self):
        self.plugin.delete_port_mac(self.context, {})
        self.driver.delete_port_mac.assert_called_once_with(self.context, {})

    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'validate_l2_gateway_for_create')
    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'create_l2_gateway')
    def test_create_l2_gateway(self, mock_create_l2gw_db,
                               mock_validate_for_create):
        fake_l2gw_id, fake_l2gw = self._get_fake_l2_gateway()
        mock_create_l2gw_db.return_value = fake_l2gw
        self.plugin.create_l2_gateway(self.context, fake_l2gw)
        mock_validate_for_create.assert_called_with(self.context,
                                                    fake_l2gw)
        mock_create_l2gw_db.assert_called_with(self.context, fake_l2gw)
        self.driver.create_l2_gateway.assert_called_once_with(self.context,
                                                              fake_l2gw)
        self.driver.create_l2_gateway_precommit.assert_called_once_with(
            self.context, fake_l2gw)
        self.driver.create_l2_gateway_postcommit.assert_called_once_with(
            self.context, fake_l2gw)

    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'validate_l2_gateway_for_delete')
    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'delete_l2_gateway')
    def test_delete_l2_gateway(self, mock_delete_l2gw_db,
                               mock_validate_for_delete):
        fake_l2gw_id, fake_l2gw = self._get_fake_l2_gateway()
        self.plugin.delete_l2_gateway(self.context, fake_l2gw_id)
        mock_validate_for_delete.assert_called_with(self.context,
                                                    fake_l2gw_id)
        mock_delete_l2gw_db.assert_called_with(self.context, fake_l2gw_id)
        self.driver.delete_l2_gateway.assert_called_once_with(self.context,
                                                              fake_l2gw_id)
        self.driver.delete_l2_gateway_precommit.assert_called_once_with(
            self.context, fake_l2gw_id)
        self.driver.delete_l2_gateway_postcommit.assert_called_once_with(
            self.context, fake_l2gw_id)

    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'validate_l2_gateway_for_update')
    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'update_l2_gateway')
    def test_update_l2_gateway(self, mock_update_l2gw_db,
                               mock_validate_for_update):
        fake_l2gw_id, fake_l2gw = self._get_fake_l2_gateway()
        mock_update_l2gw_db.return_value = fake_l2gw
        self.plugin.update_l2_gateway(self.context,
                                      fake_l2gw_id,
                                      fake_l2gw)
        mock_validate_for_update.assert_called_with(self.context,
                                                    fake_l2gw_id,
                                                    fake_l2gw)
        mock_update_l2gw_db.assert_called_with(self.context, fake_l2gw_id,
                                               fake_l2gw)
        self.driver.update_l2_gateway.assert_called_once_with(self.context,
                                                              fake_l2gw_id,
                                                              fake_l2gw)
        self.driver.update_l2_gateway_precommit.assert_called_once_with(
            self.context, fake_l2gw)
        self.driver.update_l2_gateway_postcommit.assert_called_once_with(
            self.context, fake_l2gw)

    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'validate_l2_gateway_connection_for_create')
    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'create_l2_gateway_connection')
    def test_create_l2_gateway_connection(self, mock_conn_create_l2gw_db,
                                          mock_validate_for_conn_create):
        fake_l2gw_conn_id, fake_l2gw_conn = (
            self._get_fake_l2_gateway_connection())
        mock_conn_create_l2gw_db.return_value = fake_l2gw_conn
        self.plugin.create_l2_gateway_connection(self.context,
                                                 fake_l2gw_conn)
        mock_validate_for_conn_create.assert_called_with(self.context,
                                                         fake_l2gw_conn)
        mock_conn_create_l2gw_db.assert_called_with(self.context,
                                                    fake_l2gw_conn)
        self.driver.create_l2_gateway_connection.assert_called_once_with(
            self.context, fake_l2gw_conn)
        (self.driver.create_l2_gateway_connection_precommit.
            assert_called_once_with(self.context, fake_l2gw_conn))
        (self.driver.create_l2_gateway_connection_postcommit.
            assert_called_once_with(self.context, fake_l2gw_conn))

    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'validate_l2_gateway_connection_for_delete')
    @mock.patch.object(l2gateway_db.L2GatewayMixin,
                       'delete_l2_gateway_connection')
    def test_delete_l2_gateway_connection(self, mock_conn_delete_l2gw_db,
                                          mock_validate_for_conn_delete):
        fake_l2gw_conn_id, fake_l2gw_conn = (
            self._get_fake_l2_gateway_connection())
        self.plugin.delete_l2_gateway_connection(self.context,
                                                 fake_l2gw_conn_id)
        mock_validate_for_conn_delete.assert_called_with(self.context,
                                                         fake_l2gw_conn_id)
        mock_conn_delete_l2gw_db.assert_called_with(self.context,
                                                    fake_l2gw_conn_id)
        self.driver.delete_l2_gateway_connection.assert_called_once_with(
            self.context, fake_l2gw_conn_id)
        (self.driver.delete_l2_gateway_connection_precommit.
            assert_called_once_with(self.context, fake_l2gw_conn_id))
        (self.driver.delete_l2_gateway_connection_postcommit.
            assert_called_once_with(self.context, fake_l2gw_conn_id))



# Copyright 2015 OpenStack Foundation
# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# service type constants:
L2GW = "L2GW"
l2gw = "l2gw"

AGENT_TYPE_L2GATEWAY = 'L2 Gateway agent'

L2GW_INVALID_OVSDB_IDENTIFIER = 101

ERROR_DICT = {L2GW_INVALID_OVSDB_IDENTIFIER: "Invalid ovsdb_identifier in the "
              "request"}

MONITOR = 'monitor'
OVSDB_SCHEMA_NAME = 'hardware_vtep'
OVSDB_IDENTIFIER = 'ovsdb_identifier'
L2GW_AGENT_TYPE = 'l2gw_agent_type'
NETWORK_ID = 'network_id'
SEG_ID = 'segmentation_id'
L2GATEWAY_ID = 'l2_gateway_id'
GATEWAY_RESOURCE_NAME = 'l2_gateway'
L2_GATEWAYS = 'l2-gateways'
DEVICE_ID_ATTR = 'device_name'
IFACE_NAME_ATTR = 'interfaces'
CONNECTION_RESOURCE_NAME = 'l2_gateway_connection'
EXT_ALIAS = 'l2-gateway-connection'
L2_GATEWAYS_CONNECTION = "%ss" % EXT_ALIAS
BUFFER_SIZE = 4096
MAX_RETRIES = 1000
L2_GATEWAY_SERVICE_PLUGIN = "Neutron L2 gateway Service Plugin"
PORT_FAULT_STATUS_UP = "UP"
SWITCH_FAULT_STATUS_UP = "UP"
VXLAN = "vxlan"
CREATE = "CREATE"
DELETE = "DELETE"



# Copyright (c) 2016 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from ovsdbapp.backend.ovs_idl.transaction import Transaction

from networking_l2gw.services.l2gateway.agent.ovsdb import api
from networking_l2gw.services.l2gateway.agent.ovsdb.native import (
    commands as cmd)
from networking_l2gw.services.l2gateway.agent.ovsdb.native import connection


class OvsdbHardwareVtepIdl(api.API):
    def __init__(self, context, ovsdb_conn, timeout):
        self.context = context
        self.timeout = timeout
        self.ovsdb_connection = connection.Connection(ovsdb_conn, timeout,
                                                      'hardware_vtep')
        if self.is_passive(ovsdb_conn):
            self.ovsdb_connection.accept()
        else:
            self.ovsdb_connection.start()

        self.idl = self.ovsdb_connection.idl

    @property
    def _ovs(self):
        return self._tables['Global'].rows.values()[0]

    @property
    def _tables(self):
        return self.idl.tables

    def is_passive(self, ovsdb_conn):

        return ovsdb_conn.startswith("punix:") or ovsdb_conn.startswith(
            "ptcp:")

    def transaction(self, check_error=False, log_errors=True, **kwargs):
        return Transaction(self, self.ovsdb_connection,
                           self.timeout,
                           check_error, log_errors)

    def db_find(self, table, *conditions, **kwargs):
        pass

    def get_physical_sw_list(self):
        return cmd.ListPhysicalSwitchCommand(self)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

from neutron.tests import base
from neutron_lib import rpc as n_rpc

from networking_l2gw.services.l2gateway import exceptions as l2gw_exc
from networking_l2gw.services.l2gateway.service_drivers import agent_api

import oslo_messaging as messaging


class TestL2GatewayAgentApi(base.BaseTestCase):

    def setUp(self):
        self.client_mock_p = mock.patch.object(n_rpc, 'get_client')
        self.client_mock = self.client_mock_p.start()
        self.context = mock.ANY
        self.topic = 'foo_topic'
        self.host = 'foo_host'

        self.plugin_rpc = agent_api.L2gatewayAgentApi(
            self.topic, self.host)
        super(TestL2GatewayAgentApi, self).setUp()

    def test_set_monitor_agent(self):
        cctxt = mock.Mock()
        fake_hostname = 'fake_hostname'
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.plugin_rpc.set_monitor_agent(
            self.context, fake_hostname)
        cctxt.cast.assert_called_with(
            self.context, 'set_monitor_agent',
            hostname=fake_hostname)

    def test_add_vif_to_gateway(self):
        cctxt = mock.Mock()
        fake_ovsdb_identifier = 'fake_ovsdb_id'
        fake_logical_switch = {}
        fake_physical_locator = {}
        fake_mac_remote = {}
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.plugin_rpc.add_vif_to_gateway(
            self.context, fake_ovsdb_identifier, fake_logical_switch,
            fake_physical_locator, fake_mac_remote)
        cctxt.call.assert_called_with(
            self.context, 'add_vif_to_gateway',
            ovsdb_identifier=fake_ovsdb_identifier,
            logical_switch_dict=fake_logical_switch,
            locator_dict=fake_physical_locator,
            mac_dict=fake_mac_remote)

    def test_update_vif_to_gateway(self):
        cctxt = mock.Mock()
        fake_ovsdb_identifier = 'fake_ovsdb_id'
        fake_physical_locator = {}
        fake_mac_remote = {}
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.plugin_rpc.update_vif_to_gateway(
            self.context, fake_ovsdb_identifier,
            fake_physical_locator, fake_mac_remote)
        cctxt.call.assert_called_with(
            self.context, 'update_vif_to_gateway',
            ovsdb_identifier=fake_ovsdb_identifier,
            locator_dict=fake_physical_locator,
            mac_dict=fake_mac_remote)

    def test_delete_vif_from_gateway(self):
        cctxt = mock.Mock()
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.plugin_rpc.delete_vif_from_gateway(self.context,
                                                mock.ANY,
                                                mock.ANY,
                                                mock.ANY)
        cctxt.call.assert_called_with(
            self.context, 'delete_vif_from_gateway',
            ovsdb_identifier=mock.ANY, logical_switch_uuid=mock.ANY,
            mac=mock.ANY)

    def test_delete_network(self):
        cctxt = mock.Mock()
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.plugin_rpc.delete_network(self.context, mock.ANY, mock.ANY)
        cctxt.cast.assert_called_with(
            self.context, 'delete_network', ovsdb_identifier=mock.ANY,
            logical_switch_uuid=mock.ANY)

    def test_validate_request_op_method(self):
        self.assertRaises(l2gw_exc.InvalidMethod,
                          self.plugin_rpc._validate_request_op_method,
                          self.context,
                          'fake_op_method')

    def test_update_connection_to_gateway(self):
        cctxt = mock.Mock()
        fake_ovsdb_identifier = 'fake_ovsdb_id'
        fake_logical_switch = {}
        fake_physical_locator_list = []
        fake_mac_dicts = [{}]
        fake_port_dicts = [{}]
        fake_op_method = 'DELETE'
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.plugin_rpc.update_connection_to_gateway(
            self.context, fake_ovsdb_identifier, fake_logical_switch,
            fake_physical_locator_list, fake_mac_dicts, fake_port_dicts,
            fake_op_method)
        cctxt.call.assert_called_with(
            self.context, 'update_connection_to_gateway',
            ovsdb_identifier=fake_ovsdb_identifier,
            logical_switch_dict=fake_logical_switch,
            locator_dicts=fake_physical_locator_list,
            mac_dicts=fake_mac_dicts,
            port_dicts=fake_port_dicts,
            op_method=fake_op_method)

    def test_update_connection_to_gateway_with_invalid_op_method(self):
        cctxt = mock.Mock()
        fake_ovsdb_identifier = 'fake_ovsdb_id'
        fake_logical_switch = {}
        fake_physical_locator_list = []
        fake_mac_dicts = [{}]
        fake_port_dicts = [{}]
        fake_op_method = 'create_delete_op'
        self.plugin_rpc.client.prepare.return_value = cctxt
        self.assertRaises(
            l2gw_exc.InvalidMethod,
            self.plugin_rpc.update_connection_to_gateway,
            self.context, fake_ovsdb_identifier, fake_logical_switch,
            fake_physical_locator_list, fake_mac_dicts, fake_port_dicts,
            fake_op_method)

    def test_update_connection_to_gateway_with_error(self):
        cctxt = mock.Mock()
        fake_ovsdb_identifier = 'fake_ovsdb_id'
        fake_logical_switch = {}
        fake_physical_locator_list = []
        fake_mac_dicts = [{}]
        fake_port_dicts = [{}]
        fake_op_method = 'CREATE'
        self.plugin_rpc.client.prepare.return_value = cctxt

        # Test with a timeout exception
        with mock.patch.object(cctxt,
                               'call',
                               side_effect=messaging.MessagingTimeout):
            self.assertRaises(
                l2gw_exc.OVSDBError,
                self.plugin_rpc.update_connection_to_gateway,
                self.context, fake_ovsdb_identifier, fake_logical_switch,
                fake_physical_locator_list, fake_mac_dicts, fake_port_dicts,
                fake_op_method)

        # Test with a remote exception
        with mock.patch.object(cctxt,
                               'call',
                               side_effect=Exception):
            self.assertRaises(
                l2gw_exc.OVSDBError,
                self.plugin_rpc.update_connection_to_gateway,
                self.context, fake_ovsdb_identifier, fake_logical_switch,
                fake_physical_locator_list, fake_mac_dicts, fake_port_dicts,
                fake_op_method)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.db.l2gateway.ovsdb import lib as db
from networking_l2gw.services.l2gateway.common import constants as n_const
from networking_l2gw.services.l2gateway.common import ovsdb_schema
from networking_l2gw.services.l2gateway.common import topics
from networking_l2gw.services.l2gateway.common import tunnel_calls
from networking_l2gw.services.l2gateway import exceptions as l2gw_exc
from networking_l2gw.services.l2gateway.service_drivers import agent_api

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

LOG = logging.getLogger(__name__)


class L2GatewayOVSDBCallbacks(object):
    """Implement the rpc call back functions from OVSDB."""

    target = messaging.Target(version='1.0')

    def __init__(self, plugin):
        super(L2GatewayOVSDBCallbacks, self).__init__()
        self.plugin = plugin
        self.ovsdb = None

    def update_ovsdb_changes(self, context, activity, ovsdb_data):
        """RPC to update the changes from OVSDB in the database."""
        self.ovsdb = self.get_ovsdbdata_object(
            ovsdb_data.get(n_const.OVSDB_IDENTIFIER))
        self.ovsdb.update_ovsdb_changes(context, activity, ovsdb_data)

    def notify_ovsdb_states(self, context, ovsdb_states):
        """RPC to notify the OVSDB servers connection state."""
        if ovsdb_states:
            self.ovsdb = self.get_ovsdbdata_object(
                list(ovsdb_states.keys())[0])
        if self.ovsdb:
            LOG.debug("ovsdb_states = %s", ovsdb_states)
            self.ovsdb.notify_ovsdb_states(context, ovsdb_states)

    def get_ovsdbdata_object(self, ovsdb_identifier):
        return OVSDBData(ovsdb_identifier)


class OVSDBData(object):
    """Process the data coming from OVSDB."""

    def __init__(self, ovsdb_identifier=None):
        self.ovsdb_identifier = ovsdb_identifier
        self._setup_entry_table()
        self.agent_rpc = agent_api.L2gatewayAgentApi(
            topics.L2GATEWAY_AGENT, cfg.CONF.host)
        self.l2gw_mixin = l2gateway_db.L2GatewayMixin()
        self.core_plugin = directory.get_plugin()
        self.tunnel_call = tunnel_calls.Tunnel_Calls()

    def _cleanup_all_ovsdb_tables(self, context, ovsdb_identifier):
        db.delete_all_physical_locators_by_ovsdb_identifier(
            context, ovsdb_identifier)
        db.delete_all_physical_switches_by_ovsdb_identifier(
            context, ovsdb_identifier)
        db.delete_all_physical_ports_by_ovsdb_identifier(
            context, ovsdb_identifier)
        db.delete_all_logical_switches_by_ovsdb_identifier(
            context, ovsdb_identifier)
        db.delete_all_ucast_macs_locals_by_ovsdb_identifier(
            context, ovsdb_identifier)
        db.delete_all_ucast_macs_remotes_by_ovsdb_identifier(
            context, ovsdb_identifier)
        db.delete_all_vlan_bindings_by_ovsdb_identifier(
            context, ovsdb_identifier)

    def update_ovsdb_changes(self, context, activity, ovsdb_data):
        """RPC to update the changes from OVSDB in the database."""
        ovsdb_identifier = ovsdb_data.get('ovsdb_identifier')
        if not activity:
            self._cleanup_all_ovsdb_tables(context, ovsdb_identifier)
        for item, value in ovsdb_data.items():
            lookup = self.entry_table.get(item, None)
            if lookup:
                lookup(context, value)
        if ovsdb_data.get('new_remote_macs'):
            self._handle_l2pop(context, ovsdb_data.get('new_remote_macs'))

    def notify_ovsdb_states(self, context, ovsdb_states):
        """RPC to notify the OVSDB servers connection state."""
        for ovsdb_identifier, state in ovsdb_states.items():
            if state == 'connected':
                pending_recs = db.get_all_pending_remote_macs_in_asc_order(
                    context, ovsdb_identifier)
                if pending_recs:
                    for pending_mac in pending_recs:
                        logical_switch_uuid = pending_mac['logical_switch_uuid'
                                                          ]
                        mac = pending_mac['mac']
                        operation = pending_mac['operation']
                        try:
                            if operation == 'insert' or operation == 'update':
                                l_switch = ovsdb_schema.LogicalSwitch(
                                    logical_switch_uuid, None, None, None)
                                locator_uuid = pending_mac.get(
                                    'locator_uuid', None)
                                dst_ip = pending_mac.get(
                                    'dst_ip', None)
                                locator = ovsdb_schema.PhysicalLocator(
                                    locator_uuid, dst_ip)
                                mac_remote = ovsdb_schema.UcastMacsRemote(
                                    pending_mac.get('uuid', None),
                                    mac,
                                    logical_switch_uuid,
                                    locator_uuid,
                                    pending_mac['vm_ip'])
                                if operation == 'insert':
                                    self.agent_rpc.add_vif_to_gateway(
                                        context,
                                        ovsdb_identifier,
                                        l_switch.__dict__,
                                        locator.__dict__,
                                        mac_remote.__dict__)
                                else:
                                    # update operation
                                    self.agent_rpc.update_vif_to_gateway(
                                        context,
                                        ovsdb_identifier,
                                        locator.__dict__,
                                        mac_remote.__dict__)
                            else:
                                self.agent_rpc.delete_vif_from_gateway(
                                    context, ovsdb_identifier,
                                    logical_switch_uuid, [mac])

                            # As the pending operation is over, delete the
                            # record from the pending_ucast_mac_remote table
                            db.delete_pending_ucast_mac_remote(
                                context,
                                operation,
                                ovsdb_identifier,
                                logical_switch_uuid,
                                mac)
                        except Exception as ex:
                            LOG.exception("Exception occurred = %s",
                                          str(ex))

    def _setup_entry_table(self):
        self.entry_table = {'new_logical_switches':
                            self._process_new_logical_switches,
                            'new_physical_ports':
                            self._process_new_physical_ports,
                            'new_physical_switches':
                            self._process_new_physical_switches,
                            'new_physical_locators':
                            self._process_new_physical_locators,
                            'new_local_macs':
                            self._process_new_local_macs,
                            'new_remote_macs':
                            self._process_new_remote_macs,
                            'modified_remote_macs':
                            self._process_modified_remote_macs,
                            'modified_physical_ports':
                            self._process_modified_physical_ports,
                            'deleted_logical_switches':
                            self._process_deleted_logical_switches,
                            'deleted_physical_ports':
                            self._process_deleted_physical_ports,
                            'deleted_physical_switches':
                            self._process_deleted_physical_switches,
                            'deleted_physical_locators':
                            self._process_deleted_physical_locators,
                            'deleted_local_macs':
                            self._process_deleted_local_macs,
                            'deleted_remote_macs':
                            self._process_deleted_remote_macs,
                            'modified_physical_switches':
                            self._process_modified_physical_switches,
                            }

        return

    def _process_new_logical_switches(self,
                                      context,
                                      new_logical_switches):
        for logical_switch in new_logical_switches:
            ls_dict = logical_switch
            ls_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            l_switch = db.get_logical_switch(context, ls_dict)
            if not l_switch:
                db.add_logical_switch(context, ls_dict)

    def _process_new_physical_switches(self,
                                       context,
                                       new_physical_switches):
        for physical_switch in new_physical_switches:
            ps_dict = physical_switch
            ps_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            if (ps_dict.get('tunnel_ip'))[0] == 'set':
                ps_dict['tunnel_ip'] = None
            p_switch = db.get_physical_switch(context, ps_dict)
            if not p_switch:
                db.add_physical_switch(context, ps_dict)

    def _process_new_physical_ports(self,
                                    context,
                                    new_physical_ports):
        for physical_port in new_physical_ports:
            pp_dict = physical_port
            pp_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            p_port = db.get_physical_port(context, pp_dict)
            if not p_port:
                db.add_physical_port(context, pp_dict)
            if pp_dict.get('vlan_bindings'):
                for vlan_binding in pp_dict.get('vlan_bindings'):
                    vlan_binding[
                        n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
                    vlan_binding['port_uuid'] = pp_dict.get('uuid')
                    v_binding = db.get_vlan_binding(context, vlan_binding)
                    if not v_binding:
                        db.add_vlan_binding(context, vlan_binding)

    def _process_new_physical_locators(self,
                                       context,
                                       new_physical_locators):
        for physical_locator in new_physical_locators:
            pl_dict = physical_locator
            pl_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            p_locator = db.get_physical_locator(context, pl_dict)
            if not p_locator:
                db.add_physical_locator(context, pl_dict)

    def _process_new_local_macs(self,
                                context,
                                new_local_macs):
        for local_mac in new_local_macs:
            lm_dict = local_mac
            lm_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            lm_dict['logical_switch_uuid'] = local_mac.get('logical_switch_id')
            l_mac = db.get_ucast_mac_local(context, lm_dict)
            if not l_mac:
                db.add_ucast_mac_local(context, lm_dict)

    def _process_new_remote_macs(self,
                                 context,
                                 new_remote_macs):
        for remote_mac in new_remote_macs:
            rm_dict = remote_mac
            rm_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            r_mac = db.get_ucast_mac_remote(context, rm_dict)
            if not r_mac:
                db.add_ucast_mac_remote(context, rm_dict)

    def _process_modified_remote_macs(self,
                                      context,
                                      modified_remote_macs):
        for remote_mac in modified_remote_macs:
            remote_mac[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            db.update_ucast_mac_remote(context, remote_mac)

    def _get_physical_switch_ips(self, context, mac):
        physical_switch_ips = set()
        record_dict = {n_const.OVSDB_IDENTIFIER: self.ovsdb_identifier}
        vlan_bindings = db.get_all_vlan_bindings_by_logical_switch(
            context, mac)
        for vlan_binding in vlan_bindings:
            record_dict['uuid'] = vlan_binding.get('port_uuid')
            physical_port = db.get_physical_port(context, record_dict)
            record_dict['uuid'] = physical_port.get('physical_switch_id')
            physical_switch = db.get_physical_switch(context, record_dict)
            physical_switch_ips.add(physical_switch.get('tunnel_ip'))
        return list(physical_switch_ips)

    def _get_agent_by_mac(self, context, mac):
        host = None
        mac_addr = mac.get('mac')
        port = self._get_port_by_mac(context, mac_addr)
        for port_dict in port:
            host = port_dict[portbindings.HOST_ID]
        agent_l2_pop_enabled = self._get_agent_details_by_host(context, host)
        return agent_l2_pop_enabled

    def _get_port_by_mac(self, context, mac_addr):
        port = self.core_plugin.get_ports(
            context, filters={'mac_address': [mac_addr]})
        return port

    def _get_agent_details_by_host(self, context, host):
        l2_agent = None
        agent_l2_pop_enabled = None
        agents = self.core_plugin.get_agents(
            context, filters={'host': [host]})
        for agent in agents:
            agent_tunnel_type = agent['configurations'].get('tunnel_types', [])
            agent_l2_pop_enabled = agent['configurations'].get('l2_population',
                                                               None)
            if n_const.VXLAN in agent_tunnel_type:
                l2_agent = agent
                break
        if not l2_agent:
            raise l2gw_exc.L2AgentNotFoundByHost(
                host=host)
        return agent_l2_pop_enabled

    def _handle_l2pop(self, context, new_remote_macs):
        """handle vxlan tunnel creation based on l2pop is enabled or not.

        if l2pop is enabled in L2 agent on a host to which port belongs, then
        call add_fdb_entries. otherwise, call tunnel_sync.
        """
        for mac in new_remote_macs:
            try:
                agent_l2_pop_enabled = self._get_agent_by_mac(context, mac)
            except l2gw_exc.L2AgentNotFoundByHost as e:
                LOG.debug(e.message)
                continue
            physical_switches = self._get_physical_switch_ips(context, mac)
            for physical_switch in physical_switches:
                other_fdb_entries = self._get_fdb_entries(
                    context, physical_switch, mac.get('logical_switch_id'))
                if agent_l2_pop_enabled:
                    self.tunnel_call.trigger_l2pop_sync(context,
                                                        other_fdb_entries)
                else:
                    self.tunnel_call.trigger_tunnel_sync(context,
                                                         physical_switch)

    def _process_modified_physical_ports(self,
                                         context,
                                         modified_physical_ports):
        for physical_port in modified_physical_ports:
            pp_dict = physical_port
            pp_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            modified_port = db.get_physical_port(context, pp_dict)
            if modified_port:
                db.update_physical_ports_status(context, pp_dict)
                port_vlan_bindings = physical_port.get('vlan_bindings')
                vlan_bindings = db.get_all_vlan_bindings_by_physical_port(
                    context, pp_dict)
                for vlan_binding in vlan_bindings:
                    db.delete_vlan_binding(context, vlan_binding)
                for port_vlan_binding in port_vlan_bindings:
                    port_vlan_binding['port_uuid'] = pp_dict['uuid']
                    port_vlan_binding[
                        n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
                    db.add_vlan_binding(context, port_vlan_binding)
            else:
                db.add_physical_port(context, pp_dict)

    def _process_modified_physical_switches(self, context,
                                            modified_physical_switches):
        for physical_switch in modified_physical_switches:
            db.update_physical_switch_status(context, physical_switch)

    def _process_deleted_logical_switches(self,
                                          context,
                                          deleted_logical_switches):
        for logical_switch in deleted_logical_switches:
            ls_dict = logical_switch
            ls_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            db.delete_logical_switch(context, ls_dict)

    def _process_deleted_physical_switches(self,
                                           context,
                                           deleted_physical_switches):
        for physical_switch in deleted_physical_switches:
            ps_dict = physical_switch
            ps_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            db.delete_physical_switch(context, ps_dict)
        physical_switches = db.get_all_physical_switches_by_ovsdb_id(
            context, self.ovsdb_identifier)
        if not physical_switches:
            logical_switches = db.get_all_logical_switches_by_ovsdb_id(
                context, self.ovsdb_identifier)
            if logical_switches:
                for logical_switch in logical_switches:
                    self.agent_rpc.delete_network(
                        context, self.ovsdb_identifier,
                        logical_switch.get('uuid'))

    def _process_deleted_physical_ports(self,
                                        context,
                                        deleted_physical_ports):
        for physical_port in deleted_physical_ports:
            pp_dict = physical_port
            pp_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            port_name = pp_dict['name']
            p_port = db.get_physical_port(context, pp_dict)
            if not p_port:
                raise l2gw_exc.L2GatewayInterfaceNotFound(
                    interface_id=port_name)
            p_switch_id = p_port.get('physical_switch_id')
            switch_dict = {}
            switch_dict['uuid'] = p_switch_id
            switch_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            switch_db = db.get_physical_switch(context, switch_dict)
            if not switch_db:
                raise l2gw_exc.L2GatewayDeviceNotFound(
                    device_id=p_switch_id)
            switch_name = switch_db.get('name')
            l2gw_id_list = self.l2gw_mixin._get_l2gw_ids_by_interface_switch(
                context, port_name, switch_name)
            if l2gw_id_list:
                for l2gw_id in l2gw_id_list:
                    self.l2gw_mixin._delete_connection_by_l2gw_id(context,
                                                                  l2gw_id)
            vlan_bindings = db.get_all_vlan_bindings_by_physical_port(
                context, pp_dict)
            ls_set = set()
            for vlan_binding in vlan_bindings:
                vlan_binding['logical_switch_id'] = vlan_binding.get(
                    'logical_switch_uuid')
                if vlan_binding.get('logical_switch_uuid') in ls_set:
                    db.delete_vlan_binding(context, vlan_binding)
                    continue
                bindings = db.get_all_vlan_bindings_by_logical_switch(
                    context, vlan_binding)
                if bindings and len(bindings) == 1:
                    self._delete_macs_from_ovsdb(
                        context,
                        vlan_binding.get('logical_switch_uuid'),
                        self.ovsdb_identifier)
                elif bindings and len(bindings) > 1:
                    flag = True
                    for binding in bindings:
                        if binding[
                           'ovsdb_identifier'] == self.ovsdb_identifier:
                            flag = False
                            break
                    if flag:
                        self._delete_macs_from_ovsdb(
                            context,
                            vlan_binding.get('logical_switch_uuid'),
                            self.ovsdb_identifier)
                ls_set.add(vlan_binding.get('logical_switch_uuid'))
                db.delete_vlan_binding(context, vlan_binding)
            db.delete_physical_port(context, pp_dict)

    def _delete_macs_from_ovsdb(self, context, logical_switch_id,
                                ovsdb_identifier):
        mac_list = []
        ls_dict = {'logical_switch_id': logical_switch_id,
                   'ovsdb_identifier': ovsdb_identifier}
        macs = db.get_all_ucast_mac_remote_by_ls(context, ls_dict)
        for mac in macs:
            mac_list.append(mac.get('mac'))
        self.agent_rpc.delete_vif_from_gateway(
            context, ovsdb_identifier,
            logical_switch_id, mac_list)

    def _process_deleted_physical_locators(self,
                                           context,
                                           deleted_physical_locators):
        physical_switch_ips = []
        logical_switch_ids = self._get_logical_switch_ids(context)
        physical_switches = db.get_all_physical_switches_by_ovsdb_id(
            context, self.ovsdb_identifier)
        for physical_switch in physical_switches:
            physical_switch_ips.append(
                physical_switch.get('tunnel_ip'))
        tunneling_ip_dict = self._get_agent_ips(context)
        for physical_locator in deleted_physical_locators:
            pl_dict = physical_locator
            pl_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            agent_ip = physical_locator.get('dst_ip')
            if agent_ip in tunneling_ip_dict.keys():
                for logical_switch_id in logical_switch_ids:
                    for physical_switch_ip in physical_switch_ips:
                        other_fdb_entries = self._get_fdb_entries(
                            context, physical_switch_ip, logical_switch_id)
                        agent_host = tunneling_ip_dict.get(agent_ip)
                        self.tunnel_call.trigger_l2pop_delete(
                            context, other_fdb_entries, agent_host)
            else:
                for logical_switch_id in logical_switch_ids:
                    other_fdb_entries = self._get_fdb_entries(
                        context, agent_ip, logical_switch_id)
                    self.tunnel_call.trigger_l2pop_delete(
                        context, other_fdb_entries)
            db.delete_physical_locator(context, pl_dict)

    def _process_deleted_local_macs(self,
                                    context,
                                    deleted_local_macs):
        for local_mac in deleted_local_macs:
            lm_dict = local_mac
            lm_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            db.delete_ucast_mac_local(context, lm_dict)

    def _process_deleted_remote_macs(self,
                                     context,
                                     deleted_remote_macs):
        for remote_mac in deleted_remote_macs:
            rm_dict = remote_mac
            rm_dict[n_const.OVSDB_IDENTIFIER] = self.ovsdb_identifier
            db.delete_ucast_mac_remote(context, rm_dict)

    def _get_logical_switch_ids(self, context):
        logical_switch_ids = set()
        logical_switches = db.get_all_logical_switches_by_ovsdb_id(
            context, self.ovsdb_identifier)
        for logical_switch in logical_switches:
            logical_switch_ids.add(logical_switch.get('uuid'))
        return list(logical_switch_ids)

    def _get_agent_ips(self, context):
        agent_ip_dict = {}
        agents = self.core_plugin.get_agents(
            context)
        for agent in agents:
            conf_dict_tunnel_type = agent['configurations'].get(
                "tunnel_types", [])
            if n_const.VXLAN in conf_dict_tunnel_type:
                tunnel_ip = agent['configurations'].get('tunneling_ip')
                agent_ip_dict[tunnel_ip] = agent.get('host')
        return agent_ip_dict

    def _get_fdb_entries(self, context, agent_ip, logical_switch_uuid):
        ls_dict = {'uuid': logical_switch_uuid,
                   n_const.OVSDB_IDENTIFIER: self.ovsdb_identifier}
        logical_switch = db.get_logical_switch(context, ls_dict)
        network_id = logical_switch.get('name')
        segment_id = logical_switch.get('key')
        port_fdb_entries = constants.FLOODING_ENTRY
        other_fdb_entries = {network_id: {'segment_id': segment_id,
                                          'network_type': 'vxlan',
                                          'ports': {agent_ip:
                                                    [port_fdb_entries]
                                                    }}}
        return other_fdb_entries



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import rpc as n_rpc

from networking_l2gw._i18n import _
from networking_l2gw.services.l2gateway.common import constants as n_const
from networking_l2gw.services.l2gateway import exceptions as l2gw_exc

import oslo_messaging as messaging


class L2gatewayAgentApi(object):
    """L2gateway plugin to agent RPC API."""

    API_VERSION = '1.0'

    def __init__(self, topic, host):
        """Initialize L2gateway plugin."""
        self.host = host
        target = messaging.Target(topic=topic, version=self.API_VERSION)
        self.client = n_rpc.get_client(target)

    def set_monitor_agent(self, context, hostname):
        """RPC to select Monitor/Transact agent."""
        cctxt = self.client.prepare(fanout=True)
        return cctxt.cast(context,
                          'set_monitor_agent',
                          hostname=hostname)

    def add_vif_to_gateway(self, context, ovsdb_identifier, logical_switch,
                           physical_locator, mac_remote):
        """RPC to enter the VM MAC details to gateway."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'add_vif_to_gateway',
                          ovsdb_identifier=ovsdb_identifier,
                          logical_switch_dict=logical_switch,
                          locator_dict=physical_locator,
                          mac_dict=mac_remote)

    def update_vif_to_gateway(self, context, ovsdb_identifier,
                              physical_locator, mac_remote):
        """RPC to update the VM MAC details to gateway."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'update_vif_to_gateway',
                          ovsdb_identifier=ovsdb_identifier,
                          locator_dict=physical_locator,
                          mac_dict=mac_remote)

    def delete_vif_from_gateway(self, context, ovsdb_identifier,
                                logical_switch_uuid, macs):
        """RPC to delete the VM MAC details from gateway."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'delete_vif_from_gateway',
                          ovsdb_identifier=ovsdb_identifier,
                          logical_switch_uuid=logical_switch_uuid,
                          mac=macs)

    def delete_network(self, context, ovsdb_identifier, logical_switch_uuid):
        """RPC to delete the Network from gateway."""
        cctxt = self.client.prepare()
        return cctxt.cast(context,
                          'delete_network',
                          ovsdb_identifier=ovsdb_identifier,
                          logical_switch_uuid=logical_switch_uuid)

    def _validate_request_op_method(self, context, op_method):
        """validate the method in the request."""
        method_list = [n_const.CREATE, n_const.DELETE]
        if op_method not in method_list:
            raise l2gw_exc.InvalidMethod(op_method=op_method)

    def update_connection_to_gateway(self, context, ovsdb_identifier,
                                     ls_dict, locator_list, mac_dict,
                                     port_dict, op_method):
        """RPC to update the connection to gateway."""
        self._validate_request_op_method(context, op_method)
        cctxt = self.client.prepare()
        try:
            return cctxt.call(context,
                              'update_connection_to_gateway',
                              ovsdb_identifier=ovsdb_identifier,
                              logical_switch_dict=ls_dict,
                              locator_dicts=locator_list,
                              mac_dicts=mac_dict,
                              port_dicts=port_dict,
                              op_method=op_method)
        except messaging.MessagingTimeout:
            message = _("Communication error with the L2 gateway agent")
            raise l2gw_exc.OVSDBError(message=message)
        except Exception as ex:
            message = str(ex)
            msg_splits = message.split('\n')
            raise l2gw_exc.OVSDBError(message="Error on the OVSDB "
                                      "server: " + msg_splits[0])
